"""Tests for wombat/core/data_classes.py."""

from __future__ import annotations

import datetime
from copy import deepcopy

import attr
import numpy as np
import pytest
import numpy.testing as npt

from wombat.core.library import load_yaml
from wombat.core.data_classes import (
    Failure,
    FixedCosts,
    Maintenance,
    ServiceCrew,
    StrategyMap,
    EquipmentMap,
    FromDictMixin,
    RepairRequest,
    SubassemblyData,
    ServiceEquipmentData,
    ScheduledServiceEquipmentData,
    UnscheduledServiceEquipmentData,
    valid_hour,
    convert_to_list,
    valid_reduction,
    annual_date_range,
    greater_than_zero,
    clean_string_input,
    convert_to_list_lower,
    convert_ratio_to_absolute,
)

from tests.conftest import (
    SCHEDULED_VESSEL,
    GENERATOR_SUBASSEMBLY,
    UNSCHEDULED_VESSEL_DOWNTIME,
    UNSCHEDULED_VESSEL_REQUESTS,
)


def test_convert_to_list():
    """Tests ``convert_to_list``."""
    # Tests for int conversion to list[int]
    correct_conversion = [1]
    assert convert_to_list(1) == correct_conversion

    # Tests for int conversion to list[float]
    correct_conversion = [234.56]
    assert convert_to_list(234.56) == correct_conversion

    # Tests for int conversion to list[str]
    correct_conversion = ["word"]
    assert convert_to_list("word") == correct_conversion

    # Tests for int conversion to list[str] with upper case string conversion
    correct_conversion = ["WORD"]
    assert convert_to_list("word", manipulation=str.upper) == correct_conversion


def test_clean_string_input():
    """Tests ``clean_string_input``."""
    correct = "this is a statement."
    assert clean_string_input(" THIS is a STATEMENT.    ") == correct


def test_annual_date_range_good_endpoints():
    """Tests ``annual_date_range``."""
    correct_date_range = [
        datetime.date(2019, 12, 12),
        datetime.date(2019, 12, 13),
        datetime.date(2019, 12, 14),
        datetime.date(2020, 12, 12),
        datetime.date(2020, 12, 13),
        datetime.date(2020, 12, 14),
        datetime.date(2021, 12, 12),
        datetime.date(2021, 12, 13),
        datetime.date(2021, 12, 14),
    ]
    correct_date_range = np.array(correct_date_range)
    date_range = annual_date_range(12, 14, 12, 12, 2019, 2021)
    npt.assert_equal(date_range, correct_date_range)

    # Test that bad year endpoints fail
    with pytest.raises(ValueError):
        annual_date_range(12, 14, 12, 12, 2022, 2021)

    # Test that bad month endpoints fail
    with pytest.raises(ValueError):
        annual_date_range(12, 14, 12, 1, 2019, 2021)

    # Test that bad month and day endpoints fail
    with pytest.raises(ValueError):
        annual_date_range(14, 12, 12, 12, 2019, 2021)


def test_convert_ratio_to_absolute():
    """Tests ``convert_ratio_to_absolute``."""
    # Test for <= 1
    ratio = 0.9
    total = 100
    correct_amount = 90.0
    assert convert_ratio_to_absolute(ratio, total) == correct_amount

    # Test for > 1, but small
    ratio = 1.1
    total = 100
    correct_amount = 1.1
    assert convert_ratio_to_absolute(ratio, total) == correct_amount

    # Test for large number
    ratio = 50
    total = 100
    correct_amount = 50.0
    assert convert_ratio_to_absolute(ratio, total) == correct_amount


def test_valid_hour():
    """Tests ``valid_hour``. The function being checked is an attrs validator, and is
    used in conjunction with a integer conversion, so all passed values will be forced
    to intgers, and will therefore not be checking that here.
    """

    @attr.s(auto_attribs=True)
    class HourClass:
        """Dummy class for testing ``valid_hour``."""

        hour: int = attr.ib(converter=int, validator=valid_hour)

    # Test for out of bounds below hour range.
    with pytest.raises(ValueError):
        HourClass(hour=-2)

    # Test for out of bounds above hour range.
    with pytest.raises(ValueError):
        HourClass(hour=25)

    # Test for valid null.
    hour = HourClass(hour=-1)
    assert hour.hour == -1

    # Test for valid start of bounds hour.
    hour = HourClass(hour=0)
    assert hour.hour == 0

    # Test for valid end of bounds hour.
    hour = HourClass(hour=24)
    assert hour.hour == 24


def test_valid_reduction():
    """Tests the ``valid_reduction`` validator."""

    @attr.s(auto_attribs=True)
    class ReductionClass:
        """Dummy class for testing ``valid_reduction``."""

        speed_reduction: int = attr.ib(converter=float, validator=valid_reduction)

    # Test the fringes
    with pytest.raises(ValueError):
        ReductionClass(-0.000001)
    with pytest.raises(ValueError):
        ReductionClass(-1)
    with pytest.raises(ValueError):
        ReductionClass(1.000001)
    with pytest.raises(ValueError):
        ReductionClass(2)

    # Test that valid positive decimals work
    r = ReductionClass(0.2)
    assert r.speed_reduction == 0.2

    r = ReductionClass(0.999)
    assert r.speed_reduction == 0.999


def test_greater_than_zero():
    """Tests the ``greater_than_zero`` validator."""

    @attr.s(auto_attribs=True)
    class SpeedClass:
        """Dummy class for testing ``greater_than_zero``."""

        speed: int = attr.ib(converter=float, validator=greater_than_zero)

    # Test the fringes
    with pytest.raises(ValueError):
        SpeedClass(0)
    with pytest.raises(ValueError):
        SpeedClass(-10)

    s = SpeedClass(1)
    assert s.speed == 1.0
    s = SpeedClass(20.112)
    assert s.speed == 20.112


def test_FromDictMixin():
    """Test the ``FromDictMixin`` mix in class."""

    @attr.s(auto_attribs=True)
    class DictClass(FromDictMixin):
        """A dummy class for testing the ``FromDictMixin`` methods."""

        x: int = attr.ib(converter=int)
        y: float = attr.ib(default=1, converter=float)
        z: list[str] = attr.ib(default=["empty"], converter=convert_to_list_lower)

    # Test for default values
    inputs = {"x": 2}
    cls = DictClass.from_dict(inputs)
    assert cls == DictClass(x=2)
    assert cls.x == 2
    assert cls.y == 1.0
    assert cls.z == ["empty"]

    # Test custom inputs and that kwarg overloading is allowed without storing
    # the extra inputs
    inputs = {"x": 3, "y": 3.2, "z": ["one", "two"], "arr": [3, 4, 5.5]}
    cls = DictClass.from_dict(inputs)
    assert inputs["x"] == cls.x
    assert inputs["y"] == cls.y
    assert inputs["z"] == cls.z
    assert getattr(cls, "arr", None) is None

    # Test that an error is raised if the required inputs are not passed
    with pytest.raises(AttributeError):
        DictClass.from_dict({})


def test_Maintenance():
    """Tests the `Maintenance` class."""
    # Test for all inputs being provided and converted
    inputs_all = {
        "description": "test",
        "time": 14,
        "materials": 100,
        "frequency": 200,
        "service_equipment": "ctv",
        "operation_reduction": 0.5,
        "system_value": 100000,
    }
    cls = Maintenance.from_dict(inputs_all)
    assert cls.description == "test"
    assert cls.time == 14.0
    assert cls.materials == 100.0
    assert cls.frequency == 200.0 * 24
    assert cls.service_equipment == ["CTV"]
    assert cls.operation_reduction == 0.5
    assert cls.system_value == 100000.0

    # Test for default values being populated
    inputs_defaults = {
        "time": 14,
        "materials": 100,
        "frequency": 200,
        "service_equipment": "ctv",
        "system_value": 100000,
    }
    cls = Maintenance.from_dict(inputs_defaults)
    class_data = attr.fields_dict(Maintenance)
    assert cls.description == class_data["description"].default
    assert cls.time == 14.0
    assert cls.materials == 100.0
    assert cls.frequency == 200.0 * 24
    assert cls.service_equipment == ["CTV"]
    assert cls.operation_reduction == class_data["operation_reduction"].default
    assert cls.system_value == 100000.0

    # Test for proportional materials cost, relative to system value
    inputs_system_value = {
        "description": "test",
        "time": 14,
        "materials": 0.25,
        "frequency": 200,
        "service_equipment": ["ctv", "dsv"],
        "operation_reduction": 0.5,
        "system_value": 100000,
    }
    cls = Maintenance.from_dict(inputs_system_value)
    assert cls.description == "test"
    assert cls.time == 14.0
    assert cls.materials == 25000.0
    assert cls.frequency == 200.0 * 24
    assert cls.service_equipment == ["CTV", "DSV"]
    assert cls.operation_reduction == 0.5
    assert cls.system_value == 100000.0

    # Test for assign_id
    correct_id = "M00001"
    cls.assign_id(request_id=correct_id)
    assert cls.request_id == correct_id

    # Test class is frozen by trying to set any attribute
    with pytest.raises(attr.exceptions.FrozenInstanceError):
        cls.time = 100


def test_Failure():
    """Tests the `Failure` class."""
    # Test that all inputs work
    inputs_all = {
        "scale": 0.2,
        "shape": 1.0,
        "time": 45,
        "materials": 200,
        "operation_reduction": 0.2,
        "level": 0,
        "service_equipment": "lcn",
        "system_value": 100000,
        "description": "test",
    }
    np.random.seed(0)
    cls = Failure.from_dict(inputs_all)
    assert cls.scale == 0.2
    assert cls.shape == 1.0
    assert cls.time == 45.0
    assert cls.materials == 200.0
    assert cls.operation_reduction == 0.2
    assert cls.level == 0
    assert cls.service_equipment == ["LCN"]
    assert cls.system_value == 100000
    assert cls.description == "test"
    assert cls.hours_to_next_failure() == 1394.372138301769

    # Test that the default values work
    inputs_all = {
        "scale": 0.2,
        "shape": 1.0,
        "time": 45,
        "materials": 200,
        "operation_reduction": 0.2,
        "level": 0,
        "service_equipment": "lcn",
        "system_value": 100000,
    }
    cls = Failure.from_dict(inputs_all)
    class_data = attr.fields_dict(Failure)
    assert cls.scale == 0.2
    assert cls.shape == 1.0
    assert cls.time == 45.0
    assert cls.materials == 200.0
    assert cls.operation_reduction == 0.2
    assert cls.level == 0
    assert cls.service_equipment == ["LCN"]
    assert cls.system_value == 100000
    assert cls.description == class_data["description"].default

    # Test that the proportional materials cost works
    inputs_all = {
        "scale": 0.2,
        "shape": 1.0,
        "time": 45,
        "materials": 0.01,
        "operation_reduction": 0.2,
        "level": 0,
        "service_equipment": "lcn",
        "system_value": 100000,
    }
    cls = Failure.from_dict(inputs_all)
    class_data = attr.fields_dict(Failure)
    assert cls.scale == 0.2
    assert cls.shape == 1.0
    assert cls.time == 45.0
    assert cls.materials == 1000.0
    assert cls.operation_reduction == 0.2
    assert cls.level == 0
    assert cls.service_equipment == ["LCN"]
    assert cls.system_value == 100000
    assert cls.description == class_data["description"].default

    # Test for assign_id
    correct_id = "R00001"
    cls.assign_id(request_id=correct_id)
    assert cls.request_id == correct_id

    # Test for a missing value
    inputs_missing = {
        "shape": 1.0,
        "time": 45,
        "materials": 200,
        "operation_reduction": 0.2,
        "level": 0,
        "service_equipment": "lcn",
        "system_value": 100000,
    }
    with pytest.raises(AttributeError):
        Failure.from_dict(inputs_missing)

    # Tests that with no weibull parameters, no timeouts will happen
    inputs_all = {
        "scale": 0.0,
        "shape": 0.0,
        "time": 45,
        "materials": 200,
        "operation_reduction": 0.2,
        "level": 0,
        "service_equipment": "lcn",
        "system_value": 100000,
        "description": "test",
    }
    cls = Failure.from_dict(inputs_all)
    assert cls.hours_to_next_failure() is None

    # TODO: Write a test for the weibull function


def test_SubassemblyData():
    """Tests the `SubassemblyData` class."""
    N_maintenance = len(GENERATOR_SUBASSEMBLY["maintenance"])
    failure_levels = [*GENERATOR_SUBASSEMBLY["failures"]]
    N_failure = len(failure_levels)

    subassembly = SubassemblyData.from_dict(GENERATOR_SUBASSEMBLY)
    maintenance_list = [
        Maintenance.from_dict(m) for m in GENERATOR_SUBASSEMBLY["maintenance"]
    ]
    failure_dict = {
        level: Failure.from_dict(f)
        for level, f in GENERATOR_SUBASSEMBLY["failures"].items()
    }

    assert subassembly.name == GENERATOR_SUBASSEMBLY["name"]
    assert subassembly.system_value == GENERATOR_SUBASSEMBLY["system_value"]
    assert len(subassembly.maintenance) == N_maintenance

    # Set the request ID for all maintenance tasks to avoid an AttributeError
    request = "M0000001"
    for m in maintenance_list:
        m.assign_id(request_id=request)
    for m in subassembly.maintenance:
        m.assign_id(request_id=request)
    assert subassembly.maintenance == maintenance_list
    assert len(subassembly.failures) == N_failure
    assert [*subassembly.failures] == failure_levels

    # Set the request ID for all failure tasks to avoid an AttributeError
    request = "R000000"
    for i, level in enumerate(failure_dict):
        rid = f"{request}{i}"
        failure_dict[level].assign_id(request_id=rid)
        subassembly.failures[level].assign_id(request_id=rid)
    assert subassembly.failures == failure_dict


def test_RepairRequest():
    """Tests the `RepairRequest` class."""
    failure = GENERATOR_SUBASSEMBLY["failures"][1]
    maintenance = GENERATOR_SUBASSEMBLY["maintenance"][0]

    # Test for a failure repair request
    request = {
        "system_id": "WTG-01",
        "system_name": "turbine 1",
        "subassembly_id": "generator",
        "subassembly_name": "generator",
        "severity_level": failure["level"],
        "details": Failure.from_dict(failure),
        "cable": 1,
    }
    cls = RepairRequest.from_dict(request)
    assert cls.system_id == request["system_id"]
    assert cls.system_name == request["system_name"]
    assert cls.subassembly_id == request["subassembly_id"]
    assert cls.subassembly_name == request["subassembly_name"]
    assert cls.severity_level == 1
    assert cls.cable
    assert cls.upstream_turbines == []

    # Set the id
    mid = "M00001"
    cls.assign_id(mid)
    request["details"].assign_id(mid)
    assert cls.details == request["details"]

    # Test for a maintenance repair request and default cable value
    request = {
        "system_id": "WTG-01",
        "system_name": "turbine 1",
        "subassembly_id": "generator",
        "subassembly_name": "generator",
        "severity_level": 0,
        "details": Maintenance.from_dict(maintenance),
    }
    cls = RepairRequest.from_dict(request)
    assert cls.system_id == request["system_id"]
    assert cls.system_name == request["system_name"]
    assert cls.subassembly_id == request["subassembly_id"]
    assert cls.subassembly_name == request["subassembly_name"]
    assert cls.severity_level == 0
    assert not cls.cable
    assert cls.upstream_turbines == []

    # Set the id
    mid = "R00001"
    cls.assign_id(mid)
    request["details"].assign_id(mid)
    assert cls.details == request["details"]

    # Test that details needs to be a Maintenance or Failure class
    request = {
        "system_id": "WTG-01",
        "system_name": "turbine 1",
        "subassembly_id": "generator",
        "subassembly_name": "generator",
        "severity_level": 0,
        "details": "",
    }
    with pytest.raises(TypeError):
        cls = RepairRequest.from_dict(request)


def test_ServiceCrew():
    """Tests the `ServiceCrew` class."""
    inputs = {"n_day_rate": 4, "day_rate": 100, "n_hourly_rate": 10, "hourly_rate": 6.1}
    crew = ServiceCrew.from_dict(inputs)
    assert crew.n_day_rate == 4
    assert crew.n_hourly_rate == 10
    assert crew.day_rate == 100.0
    assert crew.hourly_rate == 6.1


def test_ServiceEquipmentData_determine_type():
    """Tests the creation of the servicing equipment data classes."""
    # Test creation with requests strategy as kwarg
    vessel = ServiceEquipmentData(
        UNSCHEDULED_VESSEL_REQUESTS, strategy="requests"
    ).determine_type()
    assert isinstance(vessel, UnscheduledServiceEquipmentData)

    # Test creation with requests strategy without kwarg
    vessel = ServiceEquipmentData(UNSCHEDULED_VESSEL_REQUESTS).determine_type()
    assert isinstance(vessel, UnscheduledServiceEquipmentData)

    # Test creation with downtime strategy as kwarg
    vessel = ServiceEquipmentData(
        UNSCHEDULED_VESSEL_DOWNTIME, strategy="downtime"
    ).determine_type()
    assert isinstance(vessel, UnscheduledServiceEquipmentData)

    # Test creation with downtime strategy without kwarg
    vessel = ServiceEquipmentData(UNSCHEDULED_VESSEL_DOWNTIME).determine_type()
    assert isinstance(vessel, UnscheduledServiceEquipmentData)

    # Test creation with scheduled strategy as kwarg
    vessel = ServiceEquipmentData(
        SCHEDULED_VESSEL, strategy="scheduled"
    ).determine_type()
    assert isinstance(vessel, ScheduledServiceEquipmentData)

    # Test creation with scheduled strategy without kwarg
    vessel = ServiceEquipmentData(SCHEDULED_VESSEL).determine_type()
    assert isinstance(vessel, ScheduledServiceEquipmentData)

    # Test for error with mismatched definitions. This is a ValueError because the
    # default values are meant to cause a failure
    with pytest.raises(ValueError):
        ServiceEquipmentData(SCHEDULED_VESSEL, strategy="requests").determine_type()

    with pytest.raises(ValueError):
        ServiceEquipmentData(SCHEDULED_VESSEL, strategy="downtime").determine_type()

    with pytest.raises(ValueError):
        ServiceEquipmentData(
            UNSCHEDULED_VESSEL_REQUESTS, strategy="scheduled"
        ).determine_type()

    # Test that an invalid strategy raises a ValueError
    with pytest.raises(ValueError):
        ServiceEquipmentData(
            UNSCHEDULED_VESSEL_REQUESTS, strategy="invalid"
        ).determine_type()


def test_BaseServiceEquipmentData():
    """Tests the basic setups common to un/scheduled servicing equipment."""
    # Test that no start/end dates produce empty date sets
    vessel = ScheduledServiceEquipmentData.from_dict(SCHEDULED_VESSEL)

    vessel.set_non_operational_dates(start_year=2000, end_year=2020)
    vessel.set_reduced_speed_parameters(start_year=2000, end_year=2020)
    assert vessel.non_operational_dates.size == 0
    assert vessel.non_operational_dates_set == set()
    assert vessel.reduced_speed_dates.size == 0
    assert vessel.reduced_speed_dates_set == set()

    # Test for a single date provided that errors are raised
    data = deepcopy(SCHEDULED_VESSEL)
    data["reduced_speed_start"] = "11/14"
    with pytest.raises(ValueError):
        ScheduledServiceEquipmentData.from_dict(data)

    data["reduced_speed_start"] = None
    data["non_operational_end"] = "11/14"
    with pytest.raises(ValueError):
        ScheduledServiceEquipmentData.from_dict(data)

    # Test that same dates create an error
    data["reduced_speed_start"] = "11/14"
    data["reduced_speed_end"] = "11-14"
    with pytest.raises(ValueError):
        ScheduledServiceEquipmentData.from_dict(data)

    data["reduced_speed_start"] = None
    data["reduced_speed_end"] = None
    data["non_operational_start"] = "11-14"
    data["non_operational_end"] = "11/14"
    with pytest.raises(ValueError):
        ScheduledServiceEquipmentData.from_dict(data)

    # Test valid same-year dates
    # NOTE: 3 days x 3 years should create 9 valid dates in the set
    data = deepcopy(SCHEDULED_VESSEL)
    data["reduced_speed_start"] = "11/14"
    data["reduced_speed_end"] = "11/16"

    vessel = ScheduledServiceEquipmentData.from_dict(data)
    vessel.set_reduced_speed_parameters(start_year=2000, end_year=2002)
    assert len(vessel.reduced_speed_dates) == 9
    assert min(vessel.reduced_speed_dates) == datetime.date(2000, 11, 14)
    assert max(vessel.reduced_speed_dates) == datetime.date(2002, 11, 16)

    # Test valid overlapping year dates
    # NOTE: 4 days x 3 years, but each date chunk is split between two years
    data = deepcopy(SCHEDULED_VESSEL)
    data["non_operational_start"] = "12-30"
    data["non_operational_end"] = "1/2"
    vessel = ScheduledServiceEquipmentData.from_dict(data)
    vessel.set_non_operational_dates(start_year=2000, end_year=2002)
    assert len(vessel.non_operational_dates) == 12
    assert min(vessel.non_operational_dates) == datetime.date(2000, 1, 1)
    assert max(vessel.non_operational_dates) == datetime.date(2002, 12, 31)


def test_ScheduledServiceEquipmentData():
    """Tests the initialized attributes of ScheduledServicingEquipmentData."""
    # Test that provided values are correctly mapped
    vessel = ScheduledServiceEquipmentData.from_dict(SCHEDULED_VESSEL)
    assert vessel.name == SCHEDULED_VESSEL["name"]
    assert vessel.equipment_rate == SCHEDULED_VESSEL["equipment_rate"]
    assert vessel.n_crews == SCHEDULED_VESSEL["n_crews"]
    assert vessel.crew.day_rate == SCHEDULED_VESSEL["crew"]["day_rate"]
    assert vessel.crew.n_day_rate == SCHEDULED_VESSEL["crew"]["n_day_rate"]
    assert vessel.crew.hourly_rate == SCHEDULED_VESSEL["crew"]["hourly_rate"]
    assert vessel.crew.n_hourly_rate == SCHEDULED_VESSEL["crew"]["n_hourly_rate"]
    assert vessel.start_month == SCHEDULED_VESSEL["start_month"]
    assert vessel.start_day == SCHEDULED_VESSEL["start_day"]
    assert vessel.start_year == SCHEDULED_VESSEL["start_year"]
    assert vessel.end_month == SCHEDULED_VESSEL["end_month"]
    assert vessel.end_day == SCHEDULED_VESSEL["end_day"]
    assert vessel.end_year == SCHEDULED_VESSEL["end_year"]
    assert vessel.capability == [SCHEDULED_VESSEL["capability"].upper()]
    assert vessel.mobilization_cost == SCHEDULED_VESSEL["mobilization_cost"]
    assert vessel.mobilization_days == SCHEDULED_VESSEL["mobilization_days"]
    assert vessel.speed == SCHEDULED_VESSEL["speed"]
    assert vessel.max_windspeed_transport == SCHEDULED_VESSEL["max_windspeed_transport"]
    assert vessel.max_windspeed_repair == SCHEDULED_VESSEL["max_windspeed_repair"]
    assert (
        vessel.max_waveheight_transport == SCHEDULED_VESSEL["max_waveheight_transport"]
    )
    assert vessel.max_waveheight_repair == SCHEDULED_VESSEL["max_waveheight_repair"]
    assert vessel.workday_start == SCHEDULED_VESSEL["workday_start"]
    assert vessel.workday_end == SCHEDULED_VESSEL["workday_end"]
    assert vessel.crew_transfer_time == SCHEDULED_VESSEL["crew_transfer_time"]
    assert vessel.onsite == SCHEDULED_VESSEL["onsite"]
    assert vessel.method == SCHEDULED_VESSEL["method"].lower()
    assert vessel.operating_dates[0] == datetime.date(
        SCHEDULED_VESSEL["start_year"],
        SCHEDULED_VESSEL["start_month"],
        SCHEDULED_VESSEL["start_day"],
    )
    assert vessel.operating_dates[-1] == datetime.date(
        SCHEDULED_VESSEL["end_year"],
        SCHEDULED_VESSEL["end_month"],
        SCHEDULED_VESSEL["end_day"],
    )
    assert vessel.strategy == SCHEDULED_VESSEL["strategy"]

    # Test default values
    vessel_dict = deepcopy(SCHEDULED_VESSEL)
    has_default_value = (
        "method",
        "workday_start",
        "workday_end",
        "max_waveheight_transport",
        "max_waveheight_repair",
        "crew_transfer_time",
    )
    for parameter in has_default_value:
        vessel_dict.pop(parameter)
    vessel = ScheduledServiceEquipmentData.from_dict(vessel_dict)
    assert vessel.workday_start == vessel.workday_end == -1
    assert vessel.max_waveheight_repair == vessel.max_waveheight_transport == 1000.0
    assert vessel.method == "severity"
    assert vessel.strategy == "scheduled"
    assert vessel.crew_transfer_time == 0.0

    # Test that setting the environment's shift will work
    start_shift = 8
    end_shift = 20
    vessel._set_environment_shift(start_shift, end_shift)
    assert vessel.workday_start == start_shift
    assert vessel.workday_end == end_shift

    # Test the date range works for return visits
    vessel_dict = deepcopy(SCHEDULED_VESSEL)
    vessel_dict["onsite"] = False
    vessel_dict["start_day"] = 10
    vessel_dict["end_day"] = 11
    vessel_dict["start_month"] = 2
    vessel_dict["end_month"] = 2
    vessel_dict["start_year"] = 2019
    vessel_dict["end_year"] = 2021
    correct_date_range = np.array(
        [
            datetime.date(2019, 2, 10),
            datetime.date(2019, 2, 11),
            datetime.date(2020, 2, 10),
            datetime.date(2020, 2, 11),
            datetime.date(2021, 2, 10),
            datetime.date(2021, 2, 11),
        ]
    )
    vessel = ScheduledServiceEquipmentData.from_dict(vessel_dict)
    npt.assert_equal(vessel.operating_dates, correct_date_range)


def test_UnscheduledServiceEquipmentData():
    """Tests the initialized attributes of UnscheduledServicingEquipmentData."""
    # Test that provided values are correctly mapped for the requests-based equipment
    vessel = UnscheduledServiceEquipmentData.from_dict(UNSCHEDULED_VESSEL_REQUESTS)
    assert vessel.name == UNSCHEDULED_VESSEL_REQUESTS["name"]
    assert vessel.equipment_rate == UNSCHEDULED_VESSEL_REQUESTS["equipment_rate"]
    assert vessel.charter_days == UNSCHEDULED_VESSEL_REQUESTS["charter_days"]
    assert vessel.n_crews == UNSCHEDULED_VESSEL_REQUESTS["n_crews"]
    assert vessel.crew.day_rate == UNSCHEDULED_VESSEL_REQUESTS["crew"]["day_rate"]
    assert vessel.crew.n_day_rate == UNSCHEDULED_VESSEL_REQUESTS["crew"]["n_day_rate"]
    assert vessel.crew.hourly_rate == UNSCHEDULED_VESSEL_REQUESTS["crew"]["hourly_rate"]
    assert (
        vessel.crew.n_hourly_rate
        == UNSCHEDULED_VESSEL_REQUESTS["crew"]["n_hourly_rate"]
    )
    assert vessel.capability == [UNSCHEDULED_VESSEL_REQUESTS["capability"].upper()]
    assert vessel.mobilization_cost == UNSCHEDULED_VESSEL_REQUESTS["mobilization_cost"]
    assert vessel.mobilization_days == UNSCHEDULED_VESSEL_REQUESTS["mobilization_days"]
    assert vessel.speed == UNSCHEDULED_VESSEL_REQUESTS["speed"]
    assert vessel.strategy == UNSCHEDULED_VESSEL_REQUESTS["strategy"]
    assert (
        vessel.strategy_threshold == UNSCHEDULED_VESSEL_REQUESTS["strategy_threshold"]
    )
    assert (
        vessel.max_windspeed_transport
        == UNSCHEDULED_VESSEL_REQUESTS["max_windspeed_transport"]
    )
    assert (
        vessel.max_windspeed_repair
        == UNSCHEDULED_VESSEL_REQUESTS["max_windspeed_repair"]
    )
    assert (
        vessel.max_waveheight_transport
        == UNSCHEDULED_VESSEL_REQUESTS["max_waveheight_transport"]
    )
    assert (
        vessel.max_waveheight_repair
        == UNSCHEDULED_VESSEL_REQUESTS["max_waveheight_repair"]
    )
    assert vessel.workday_start == UNSCHEDULED_VESSEL_REQUESTS["workday_start"]
    assert vessel.workday_end == UNSCHEDULED_VESSEL_REQUESTS["workday_end"]
    assert (
        vessel.crew_transfer_time == UNSCHEDULED_VESSEL_REQUESTS["crew_transfer_time"]
    )
    assert vessel.onsite == UNSCHEDULED_VESSEL_REQUESTS["onsite"]
    assert vessel.method == UNSCHEDULED_VESSEL_REQUESTS["method"].lower()

    # Test default values
    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_REQUESTS)
    has_default_value = (
        "method",
        "workday_start",
        "workday_end",
        "max_waveheight_transport",
        "max_waveheight_repair",
        "crew_transfer_time",
    )
    for parameter in has_default_value:
        vessel_dict.pop(parameter)
    vessel = UnscheduledServiceEquipmentData.from_dict(vessel_dict)
    assert vessel.workday_start == vessel.workday_end == -1
    assert vessel.max_waveheight_repair == vessel.max_waveheight_transport == 1000.0
    assert vessel.method == "severity"
    assert vessel.crew_transfer_time == 0.0

    # Test that setting the environment's shift will work
    start_shift = 8
    end_shift = 20
    vessel._set_environment_shift(start_shift, end_shift)
    assert vessel.workday_start == start_shift
    assert vessel.workday_end == end_shift

    # Test the strategy and threshold bounds with edge of in bounds for requests
    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_REQUESTS)
    vessel_dict["strategy_threshold"] = 1
    vessel = UnscheduledServiceEquipmentData.from_dict(vessel_dict)
    assert vessel.strategy_threshold == 1

    # Test for edge of out of bounds for requests
    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_REQUESTS)
    vessel_dict["strategy_threshold"] = 0
    with pytest.raises(ValueError):
        UnscheduledServiceEquipmentData.from_dict(vessel_dict)

    # Test the strategy and threshold bounds with edge of in bounds for downtime
    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_DOWNTIME)
    vessel_dict["strategy_threshold"] = 0.0000000000000001
    vessel = UnscheduledServiceEquipmentData.from_dict(vessel_dict)
    assert vessel.strategy_threshold == 0.0000000000000001

    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_DOWNTIME)
    vessel_dict["strategy_threshold"] = 0.99999999999999
    vessel = UnscheduledServiceEquipmentData.from_dict(vessel_dict)
    assert vessel.strategy_threshold == 0.99999999999999

    # Test the strategy and threshold bounds with edge for out of bounds for downtime
    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_DOWNTIME)
    vessel_dict["strategy_threshold"] = 0
    with pytest.raises(ValueError):
        UnscheduledServiceEquipmentData.from_dict(vessel_dict)

    vessel_dict = deepcopy(UNSCHEDULED_VESSEL_DOWNTIME)
    vessel_dict["strategy_threshold"] = 1
    with pytest.raises(ValueError):
        UnscheduledServiceEquipmentData.from_dict(vessel_dict)

    # Test that setting the environment's shift will work
    start_shift = 8
    end_shift = 20
    vessel._set_environment_shift(start_shift, end_shift)
    assert vessel.workday_start == start_shift
    assert vessel.workday_end == end_shift


def test_FixedCosts_operations_provided():
    """Tests high resolution inptus go to 0 when `operations` is provided."""
    high_res = FixedCosts(
        operations=100,  # should be 100
        operations_management_administration=100,
        project_management_administration=100,
        marine_management=111,
        weather_forecasting=22,
        condition_monitoring=47,
        operating_facilities=25,
        environmental_health_safety_monitoring=45,
        insurance=100,
        brokers_fee=44,
        operations_all_risk=100,
        business_interruption=76,
        third_party_liability=82,
        storm_coverage=98,
        annual_leases_fees=100,
        submerge_land_lease_costs=80,
        transmission_charges_rights=90,
        onshore_electrical_maintenance=60,
        labor=100,
    )

    assert high_res.operations == 100
    assert high_res.operations_management_administration == 0
    assert high_res.project_management_administration == 0
    assert high_res.marine_management == 0
    assert high_res.weather_forecasting == 0
    assert high_res.condition_monitoring == 0
    assert high_res.operating_facilities == 0
    assert high_res.environmental_health_safety_monitoring == 0
    assert high_res.insurance == 0
    assert high_res.brokers_fee == 0
    assert high_res.operations_all_risk == 0
    assert high_res.business_interruption == 0
    assert high_res.third_party_liability == 0
    assert high_res.storm_coverage == 0
    assert high_res.annual_leases_fees == 0
    assert high_res.submerge_land_lease_costs == 0
    assert high_res.transmission_charges_rights == 0
    assert high_res.onshore_electrical_maintenance == 0
    assert high_res.labor == 0


def test_FixedCosts_category_values_provided():
    """Test that High resolution inputs get overwritten in the case of providing
    category values.
    """
    high_res = FixedCosts(
        operations=0,  # should be 530
        operations_management_administration=100,
        project_management_administration=100,
        marine_management=111,
        weather_forecasting=22,
        condition_monitoring=47,
        operating_facilities=25,  # independent
        environmental_health_safety_monitoring=45,  # independent
        insurance=100,
        brokers_fee=44,
        operations_all_risk=100,
        business_interruption=76,
        third_party_liability=82,
        storm_coverage=98,
        annual_leases_fees=100,
        submerge_land_lease_costs=80,
        transmission_charges_rights=90,
        onshore_electrical_maintenance=60,  # independent
        labor=100,  # independent
    )
    assert high_res.operations == 530
    assert high_res.operations_management_administration == 100
    assert high_res.project_management_administration == 0
    assert high_res.marine_management == 0
    assert high_res.weather_forecasting == 0
    assert high_res.condition_monitoring == 0
    assert high_res.operating_facilities == 25
    assert high_res.environmental_health_safety_monitoring == 45
    assert high_res.insurance == 100
    assert high_res.brokers_fee == 0
    assert high_res.operations_all_risk == 0
    assert high_res.business_interruption == 0
    assert high_res.third_party_liability == 0
    assert high_res.storm_coverage == 0
    assert high_res.annual_leases_fees == 100
    assert high_res.submerge_land_lease_costs == 0
    assert high_res.transmission_charges_rights == 0
    assert high_res.onshore_electrical_maintenance == 60
    assert high_res.labor == 100


def test_equipment_map(env_setup):
    """Tests the ``EquipmentMap`` class."""
    env = env_setup

    # Test for a requests-based equipment
    hlv = ServiceEquipmentData(
        load_yaml(env.data_dir / "vessels", "hlv_requests.yaml")
    ).determine_type()
    mapping = EquipmentMap(equipment=hlv, strategy_threshold=hlv.strategy_threshold)
    assert mapping.equipment == hlv
    assert mapping.strategy_threshold == hlv.strategy_threshold

    # Test for a downtime-based equipment
    hlv = ServiceEquipmentData(
        load_yaml(env.data_dir / "vessels", "hlv_downtime.yaml")
    ).determine_type()
    mapping = EquipmentMap(equipment=hlv, strategy_threshold=hlv.strategy_threshold)
    assert mapping.equipment == hlv
    assert mapping.strategy_threshold == hlv.strategy_threshold


def test_strategy_map(env_setup):
    """Tests the ``StrategyMap`` class."""
    env = env_setup

    mapping = StrategyMap()
    assert mapping.CTV == []
    assert mapping.SCN == []
    assert mapping.LCN == []
    assert mapping.CAB == []
    assert mapping.RMT == []
    assert mapping.DRN == []
    assert mapping.DSV == []
    assert not mapping.is_running

    # Test for bad input
    hlv = load_yaml(env.data_dir / "vessels", "hlv_requests.yaml")
    hlv["capability"] = ["SCN", "LCN"]
    hlv = ServiceEquipmentData(hlv).determine_type()
    hlv_map = EquipmentMap(hlv.strategy_threshold, hlv)

    # List of inputs will fail
    with pytest.raises(ValueError):
        mapping.update(hlv.capability, hlv.strategy_threshold, hlv)

    # Single bad input will fail
    with pytest.raises(ValueError):
        mapping.update("CRV", hlv.strategy_threshold, hlv)

    # Check that the update did not complete
    assert not mapping.is_running

    # Create a few equipment and test for inclusion

    # HLV with multiple capabilities
    hlv = load_yaml(env.data_dir / "vessels", "hlv_requests.yaml")
    hlv["capability"] = ["SCN", "LCN"]
    hlv = ServiceEquipmentData(hlv).determine_type()
    hlv_map = EquipmentMap(hlv.strategy_threshold, hlv)
    for capability in hlv.capability:
        mapping.update(capability, hlv.strategy_threshold, hlv)
    assert mapping.SCN == mapping.LCN == [hlv_map]
    assert mapping.is_running

    # Cabling vessel with one capability
    cab = load_yaml(env.data_dir / "vessels", "cabling.yaml")
    cab = ServiceEquipmentData(cab).determine_type()
    for capability in cab.capability:
        mapping.update(capability, cab.strategy_threshold, cab)
    cab_map = EquipmentMap(cab.strategy_threshold, cab)
    assert mapping.CAB == [cab_map]
    assert mapping.is_running

    # Diving support vessel
    dsv = load_yaml(env.data_dir / "vessels", "dsv.yaml")
    dsv = ServiceEquipmentData(dsv).determine_type()
    for capability in dsv.capability:
        mapping.update(capability, dsv.strategy_threshold, dsv)
    dsv_map = EquipmentMap(dsv.strategy_threshold, dsv)
    assert mapping.DSV == [dsv_map]
    assert mapping.is_running

    # Remote reset
    rmt = load_yaml(env.data_dir / "vessels", "rmt.yaml")
    rmt = ServiceEquipmentData(rmt).determine_type()
    for capability in rmt.capability:
        mapping.update(capability, rmt.strategy_threshold, rmt)
    rmt_map = EquipmentMap(rmt.strategy_threshold, rmt)
    assert mapping.RMT == [rmt_map]
    assert mapping.is_running

    # Drone repair
    drn = load_yaml(env.data_dir / "vessels", "drn.yaml")
    drn = ServiceEquipmentData(drn).determine_type()
    for capability in drn.capability:
        mapping.update(capability, drn.strategy_threshold, drn)
    drn_map = EquipmentMap(drn.strategy_threshold, drn)
    assert mapping.DRN == [drn_map]
    assert mapping.is_running

    # Crew tranfer
    ctv = load_yaml(env.data_dir / "vessels", "ctv_requests.yaml")
    ctv = ServiceEquipmentData(ctv).determine_type()
    for capability in ctv.capability:
        mapping.update(capability, ctv.strategy_threshold, ctv)
    ctv_map = EquipmentMap(ctv.strategy_threshold, ctv)
    assert mapping.CTV == [ctv_map]
    assert mapping.is_running

    # Update an existing category with another piece of equipment
    fsv = load_yaml(env.data_dir / "vessels", "fsv_downtime.yaml")
    fsv = ServiceEquipmentData(fsv).determine_type()
    for capability in fsv.capability:
        mapping.update(capability, fsv.strategy_threshold, fsv)
    fsv_map = EquipmentMap(fsv.strategy_threshold, fsv)
    assert mapping.SCN == [hlv_map, fsv_map]
    assert mapping.is_running


def test_FixedCosts_high_resolution_provided():
    """Test that high resolution inputs work and categories sum correctly."""
    high_res = FixedCosts(
        operations=0,  # should be 1080
        operations_management_administration=0,  # should be 280
        project_management_administration=100,
        marine_management=111,
        weather_forecasting=22,
        condition_monitoring=47,
        operating_facilities=25,  # independent
        environmental_health_safety_monitoring=45,  # independent
        insurance=0,  # should be 400
        brokers_fee=44,
        operations_all_risk=100,
        business_interruption=76,
        third_party_liability=82,
        storm_coverage=98,
        annual_leases_fees=0,  # should be 170
        submerge_land_lease_costs=80,
        transmission_charges_rights=90,
        onshore_electrical_maintenance=60,  # independent
        labor=100,  # independent
    )
    assert high_res.operations == 1080
    assert high_res.operations_management_administration == 280
    assert high_res.insurance == 400
    assert high_res.annual_leases_fees == 170



"""Tests wombat/core/environment.py."""


from __future__ import annotations

import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import pytest
import numpy.testing as npt
from pandas.testing import assert_index_equal

from wombat.core import RepairManager, WombatEnvironment
from wombat.windfarm import Windfarm

from tests.conftest import TEST_DATA


def test_data_dir():
    """Tests the data_dir creation."""
    relative_library = Path(__file__).resolve().parent / "library"
    assert TEST_DATA == relative_library


def test_setup():
    """Tests the setup of the `WombatEnvironment` and logging infrastructure."""
    # Test an invalid directory
    with pytest.raises(FileNotFoundError):
        WombatEnvironment(
            data_dir="TEST_DATA",
            weather_file="test_weather_quick_load.csv",
            workday_start=8,
            workday_end=18,
        )

    # Test invalid starting hour - too low
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather_quick_load.csv",
            workday_start=-1,
            workday_end=18,
        )

    # Test invalid starting hour - too high
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather_quick_load.csv",
            workday_start=24,
            workday_end=18,
        )

    # Test invalid ending hour - too low
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather_quick_load.csv",
            workday_start=8,
            workday_end=-1,
        )

    # Test invalid ending hour - before start
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather_quick_load.csv",
            workday_start=8,
            workday_end=3,
        )

    # Test invalid ending hour - at the same time as start
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather_quick_load.csv",
            workday_start=8,
            workday_end=8,
        )

    # Test invalid start_year - later than file limits (2002 through 2003)
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather.csv",
            workday_start=8,
            workday_end=16,
            simulation_name="testing_setup",
            start_year=2004,
            end_year=2003,
        )

    # Test invalid end_year - start year is after end year
    with pytest.raises(ValueError):
        WombatEnvironment(
            data_dir=TEST_DATA,
            weather_file="test_weather.csv",
            workday_start=8,
            workday_end=16,
            simulation_name="testing_setup",
            start_year=2003,
            end_year=2002,
        )

    # Test the data are read correctly with default starting and ending year
    # and the logging infrastructure
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    assert env.workday_start == 8
    assert env.workday_end == 16
    assert env.shift_length == 8
    assert isinstance(env.weather, pd.DataFrame)
    assert env.start_datetime == datetime.datetime(2002, 1, 1, 0, 0)
    assert env.end_datetime == datetime.datetime(2003, 12, 31, 23, 0)
    assert env.max_run_time == 8760 * 2  # 2 year data file

    assert Path(env.events_log_fname).is_file()
    assert Path(env.operations_log_fname).is_file()

    # Delete the logged files, and check they not longer exist
    env.cleanup_log_files()
    assert not Path(env.events_log_fname).is_file()
    assert not Path(env.operations_log_fname).is_file()

    # Test custom start year
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
        start_year=2003,
    )
    assert env.start_datetime == datetime.datetime(2003, 1, 1, 0, 0)
    assert env.end_datetime == datetime.datetime(2003, 12, 31, 23, 0)
    env.cleanup_log_files()  # delete the logged data

    # Test custom end year
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
        end_year=2002,
    )
    assert env.start_datetime == datetime.datetime(2002, 1, 1, 0, 0)
    assert env.end_datetime == datetime.datetime(2002, 12, 31, 23, 0)
    env.cleanup_log_files()  # delete the logged data

    # Test custom start and end year
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
        start_year=2003,
        end_year=2003,
    )
    assert env.start_datetime == datetime.datetime(2003, 1, 1, 0, 0)
    assert env.end_datetime == datetime.datetime(2003, 12, 31, 23, 0)
    env.cleanup_log_files()  # delete the logged data


def test_timing():
    """Test basic movements and time coordination for `WombatEnvironment`.

    Methods that are completely tested:
     - ``date_ix``
     - ``simulation_time``
     - ``weather_now``

    Methods that are partially tested for no inputs:
     - ``hours_to_next_shift``
     - ``is_workshift``

    """
    # Setup a basic environment
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather_quick_load.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    manager = RepairManager(env)
    Windfarm(env, "layout.csv", manager)

    # Run for 96 hours from the start to this point in the data
    # 1/5/02 0:00,8.238614098,0.37854216
    correct_index = 96
    correct_date = datetime.date(2002, 1, 5)
    correct_hour = 0
    correct_datetime = datetime.datetime(2002, 1, 5, 0, 0, 0)
    correct_wind = 8.238614098
    correct_wave = 0.37854216

    assert env.date_ix(correct_date) == correct_index
    assert env.date_ix(correct_datetime) == correct_index

    env.run(until=correct_index)
    assert env.now == correct_index
    assert env.simulation_time == correct_datetime
    assert env.hours_to_next_shift() == 8
    assert all(env.weather_now == (correct_wind, correct_wave, correct_hour))
    assert not env.is_workshift()

    # Test for a non-even timing of 128.7 hours
    # 1/6/02 8:00,8.321216829,0.795805105
    until = 128.7
    correct_index = 120
    correct_date = datetime.date(2002, 1, 6)
    correct_datetime = datetime.datetime(2002, 1, 6, 8, 42)
    correct_hours_to_next_shift = 23.3
    correct_hour = 8
    correct_wind = 8.321216829
    correct_wave = 0.795805105
    assert env.date_ix(correct_date) == correct_index
    assert env.date_ix(correct_datetime) == correct_index

    env.run(until=until)
    assert env.now == until
    assert env.simulation_time == correct_datetime
    assert env.hours_to_next_shift() == correct_hours_to_next_shift
    assert all(env.weather_now == (correct_wind, correct_wave, correct_hour))
    assert env.is_workshift()

    env.cleanup_log_files()  # delete the logged data


def test_is_workshift():
    """Tests the ``is_workshift`` method for a variety of inputs."""
    # Setup a basic environment
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather_quick_load.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    manager = RepairManager(env)
    Windfarm(env, "layout.csv", manager)
    # Test for the 24 hours of the day with a basic 8am to 4pm workday
    assert not env.is_workshift()
    for i in range(23):
        env.run(i + 1)
        if 7 <= i < 15:
            # THe hours from 8am until, but not at, 4pm are working hours
            assert env.is_workshift()
        else:
            assert not env.is_workshift()

        # Test the boundary conditions

        # Start and current are the same time, but end is midnight
        assert env.is_workshift(workday_start=i + 1, workday_end=24)

        # Start is just before current, but the end is midnight
        assert env.is_workshift(workday_start=i, workday_end=24)

        # Start is just after, but the end is midnight
        assert not env.is_workshift(workday_start=i + 2, workday_end=24)

        # Start is before current, but end is the current hour
        assert not env.is_workshift(workday_start=i, workday_end=i + 1)

    env.cleanup_log_files()  # delete the logged data


def test_hour_in_shift():
    """Tests the ``hour_in_shift`` method for a variety of inputs."""
    # Setup a basic environment
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather_quick_load.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    manager = RepairManager(env)
    Windfarm(env, "layout.csv", manager)

    # Test the boundary conditions at midnight

    i = 0
    # Start and current are the same time, but end is midnight
    assert env.hour_in_shift(hour=i, workday_start=i, workday_end=24)

    # Start is just before current (-1), so reverts to env workday settings
    assert not env.hour_in_shift(hour=i, workday_start=-1, workday_end=24)

    # Start is just after, but the end is midnight
    assert not env.hour_in_shift(hour=i, workday_start=i + 1, workday_end=24)

    # Start is before current, but end is the current hour
    assert not env.hour_in_shift(hour=i, workday_start=i, workday_end=i)

    # Test for the 24 hours of the day, post midnight
    for i in range(1, 24):
        # Start and current are the same time, but end is midnight
        assert env.hour_in_shift(hour=i, workday_start=i, workday_end=24)

        # Start is just before current, but the end is midnight
        assert env.hour_in_shift(hour=i, workday_start=i - 1, workday_end=24)

        # Start is just after, but the end is midnight
        assert not env.hour_in_shift(hour=i, workday_start=i + 1, workday_end=24)

        # Start is before current, but end is the current hour
        assert not env.hour_in_shift(hour=i, workday_start=i, workday_end=i)

    env.cleanup_log_files()  # delete the logged data


def test_hours_to_next_shift():
    """Tests the ``hours_to_next_shift`` method for a variety of inputs."""
    # Setup a basic environment
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather_quick_load.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    manager = RepairManager(env)
    Windfarm(env, "layout.csv", manager)

    # Test that at hour 0, there are 8 hours until 8am
    assert env.hours_to_next_shift() == 8

    # Test that at hour 0, with start time 0, there are 24 hours until the next period
    assert env.hours_to_next_shift(workday_start=0)

    env.run(until=7)

    # Test that there is still 1 hour left
    assert env.hours_to_next_shift() == 1

    # Test if the starting shift is at the same time, the we get 24 hours
    assert env.hours_to_next_shift(workday_start=7) == 24

    # Test that it's 24 hours left
    env.run(until=8)
    assert env.hours_to_next_shift() == 24

    # Test for an uneven time until the next shift
    env.run(until=9.3)
    assert env.hours_to_next_shift() == pytest.approx(22.7, rel=1e-4)  # weird floats

    env.cleanup_log_files()  # delete the logged data


def test_weather_forecast():
    """Tests the ``weather_forecast`` method for a variety of inputs."""
    # Setup a basic environment
    env = WombatEnvironment(
        data_dir=TEST_DATA,
        weather_file="test_weather_quick_load.csv",
        workday_start=8,
        workday_end=16,
        simulation_name="testing_setup",
    )
    manager = RepairManager(env)
    Windfarm(env, "layout.csv", manager)

    # Test for the next 5 hours, but at the top of the hour
    # Starts with hour 0, and ends with the 5th hour following
    correct_index = pd.DatetimeIndex(
        pd.date_range("1/1/2002 00:00:00", "1/1/2002 04:00:00", freq="1H"),
        name="datetime",
    )
    correct_hour = np.array([0, 1, 2, 3, 4], dtype=float)
    correct_wind = np.array(
        [11.75561096, 10.41321252, 8.959270788, 9.10014808, 9.945059601]
    )
    correct_wave = np.array(
        [1.281772405, 1.586584315, 1.725690828, 1.680982063, 1.552232138]
    )

    print(env.weather.shape)
    ix, hour, wind, wave = env.weather_forecast(hours=5)
    assert ix.size == wind.size == wave.size == 5
    assert_index_equal(ix, correct_index)
    npt.assert_equal(hour, correct_hour)
    npt.assert_equal(wind, correct_wind)
    npt.assert_equal(wave, correct_wave)

    # Test for 5 hours at an uneven start time increment
    # Starts at hour 1, and ends with the 5th hour following
    env.run(until=0.1)
    correct_index = pd.DatetimeIndex(
        pd.date_range("1/1/2002 00:00:00", "1/1/2002 05:00:00", freq="1H"),
        name="datetime",
    )
    correct_hour = np.arange(6, dtype=float)
    correct_wind = np.array(
        [11.75561096, 10.41321252, 8.959270788, 9.10014808, 9.945059601, 11.50807971]
    )
    correct_wave = np.array(
        [1.281772405, 1.586584315, 1.725690828, 1.680982063, 1.552232138, 1.335083363]
    )

    ix, hour, wind, wave = env.weather_forecast(hours=5)
    assert ix.size == hour.size == wind.size == wave.size == 6
    assert_index_equal(ix, correct_index)
    npt.assert_equal(hour, correct_hour)
    npt.assert_equal(wind, correct_wind)
    npt.assert_equal(wave, correct_wave)

    # Test for 5.5 hours at an uneven start time increment
    # Starts at hour 1, and ends at the 7th hour following
    env.run(until=1.2)
    correct_index = pd.DatetimeIndex(
        pd.date_range("1/1/2002 01:00:00", "1/1/2002 07:00:00", freq="1H"),
        name="datetime",
    )
    correct_hour = np.arange(1, 8, dtype=float)
    correct_wind = np.array(
        [
            10.41321252,
            8.959270788,
            9.10014808,
            9.945059601,
            11.50807971,
            12.44757098,
            13.33747206,
        ]
    )
    correct_wave = np.array(
        [
            1.586584315,
            1.725690828,
            1.680982063,
            1.552232138,
            1.335083363,
            1.294392727,
            1.187261471,
        ]
    )

    ix, hour, wind, wave = env.weather_forecast(hours=5.5)
    assert ix.size == hour.size == wind.size == wave.size == 7
    assert_index_equal(ix, correct_index)
    npt.assert_equal(hour, correct_hour)
    npt.assert_equal(wind, correct_wind)
    npt.assert_equal(wave, correct_wave)

    env.cleanup_log_files()  # delete the logged data



"""The postprocessing metric computation."""
from __future__ import annotations

import logging
import warnings
from copy import deepcopy
from math import fsum
from pathlib import Path
from itertools import product

import numpy as np
import PySAM
import pandas as pd
import PySAM.PySSC as pssc
import PySAM.Singleowner as pysam_singleowner_financial_model

from wombat.core import FixedCosts
from wombat.core.library import load_yaml


def _check_frequency(frequency: str, which: str = "all") -> str:
    """Checks the frequency input to ensure it meets the correct criteria according
    to the ``which`` flag.

    Parameters
    ----------
    frequency : str
        The user-provided value.
    which : str, optional
        Designation for which combinations to check for, by default "all".
        - "all": project, annual, monthly, and month-year

    Returns
    -------
    str
        The lower-case, input with white spaces removed.

    Raises
    ------
    ValueError
        Raised if an invalid value was raised
    """
    if which == "all":
        opts = ("project", "annual", "monthly", "month-year")  # type: ignore
    elif which == "monthly":
        opts = ("project", "annual", "monthly")  # type: ignore
    elif which == "annual":  # type: ignore
        opts = ("project", "annual")  # type: ignore
    frequency = frequency.lower().strip()  # type: ignore
    if frequency not in opts:
        raise ValueError(f"``frequency`` must be one of {opts}.")  # type: ignore
    return frequency


def _calculate_time_availability(
    availability: np.ndarray, by_turbine: bool = False
) -> float | np.ndarray:
    """Calculates the availability ratio of the whole timeseries or the whole
    timeseries, by turbine.

    Parameters
    ----------
    availability : np.ndarray
        Timeseries array of operating ratios.
    by_turbine : bool, optional
        If True, calculates the availability rate of each column, otherwise across the
        whole array, by default False.

    Returns
    -------
    float | np.ndarray
        Availability ratio across the whole timeseries, or broken out by column
        (turbine).
    """
    availability = availability > 0
    if by_turbine:
        return availability.sum(axis=0) / availability.shape[0]
    return availability.sum() / availability.size


class Metrics:
    """The metric computation class for storing logs and compiling results."""

    _hourly_cost = "hourly_labor_cost"
    _salary_cost = "salary_labor_cost"
    _labor_cost = "total_labor_cost"
    _equipment_cost = "equipment_cost"
    _materials_cost = "materials_cost"
    _total_cost = "total_cost"
    _cost_columns = [
        _hourly_cost,
        _salary_cost,
        _labor_cost,
        _equipment_cost,
        _materials_cost,
        _total_cost,
    ]

    def __init__(
        self,
        data_dir: str | Path,
        events: str | pd.DataFrame,
        operations: str | pd.DataFrame,
        potential: str | pd.DataFrame,
        production: str | pd.DataFrame,
        inflation_rate: float,
        project_capacity: float,
        turbine_capacities: list[float],
        substation_id: str | list[str],
        turbine_id: str | list[str],
        substation_turbine_map: dict[str, dict[str, list[str]]],
        service_equipment_names: str | list[str],
        fixed_costs: str | None = None,
        SAM_settings: str | None = None,
    ) -> None:
        """Initializes the Metrics class.

        Parameters
        ----------
        data_dir : str | Path
            This should be the same as was used for running the analysis.
        events : str | pd.DataFrame
            Either a pandas ``DataFrame`` or filename to be used to read the csv log
            data.
        operations : str | pd.DataFrame
            Either a pandas ``DataFrame`` or filename to be used to read the csv log
            data.
        potential : str | pd.DataFrame
            Either a pandas ``DataFrame`` or a filename to be used to read the csv
            potential power production data.
        production : str | pd.DataFrame
            Either a pandas ``DataFrame`` or a filename to be used to read the csv power
            production data.
        inflation_rate : float
            The inflation rate to be applied to all dollar amounts from the analysis
            starting year to ending year.
        project_capacity : float
            The project's rated capacity, in MW.
        turbine_capacities : Union[float, List[float]]
            The capacity of each individual turbine corresponding to ``turbine_id``, in
            kW.
        substation_id : str | list[str]
            The substation id(s).
        turbine_id : str | list[str]
            The turbine id(s).
        substation_turbine_map : dict[str, dict[str, list[str]]]
            A copy of ``Windfarm.substation_turbine_map``. This is a dictionary mapping
            of the subation IDs (keys) and a nested dictionary of its associated turbine
            IDs and each turbine's total plant weighting (turbine capacity / plant
            capacity).
        service_equipment_names : str | list[str]
            The names of the servicing equipment, corresponding to
            ``ServiceEquipment.settings.name`` for each ``ServiceEquipment`` in the
            simulation.
        fixed_costs : str | None
            The filename of the project's fixed costs.
        SAM_settings : str | None
            The SAM settings YAML file located in <data_dir>/windfarm/<SAM_settings>
            that should end in ".yaml". If no input is provided, then the model will
            raise a ``NotImplementedError`` when the SAM-powered metrics are attempted
            to be accessed.

            .. warning:: This functionality relies heavily on the user to configure
                correctly. More information can be found at:
                https://nrel-pysam.readthedocs.io/en/master/modules/Singleowner.html
        """
        self.data_dir = Path(data_dir)
        if not self.data_dir.is_dir():
            raise FileNotFoundError(f"{self.data_dir} does not exist")

        self.inflation_rate = 1 + inflation_rate
        self.project_capacity = project_capacity

        if fixed_costs is None:
            # Create a zero-cost FixedCosts object
            self.fixed_costs = FixedCosts.from_dict({"operations": 0})  # type: ignore
        else:
            try:
                assert isinstance(fixed_costs, str)
                fixed_costs = load_yaml(self.data_dir / "project/config", fixed_costs)
            except FileNotFoundError:
                assert isinstance(fixed_costs, str)
                fixed_costs = load_yaml(self.data_dir / "windfarm", fixed_costs)
                logging.warning(
                    "DeprecationWarning: In v0.8, all fixed cost configurations must be"
                    " located in: '<library>/project/config/"
                )
            self.fixed_costs = FixedCosts.from_dict(fixed_costs)  # type: ignore

        if isinstance(substation_id, str):
            substation_id = [substation_id]
        self.substation_id = substation_id

        if isinstance(turbine_id, str):
            turbine_id = [turbine_id]
        self.turbine_id = turbine_id

        self.substation_turbine_map = substation_turbine_map
        self.turbine_weights = (
            pd.concat([pd.DataFrame(val) for val in substation_turbine_map.values()])
            .set_index("turbines")
            .T
        )

        if isinstance(service_equipment_names, str):
            service_equipment_names = [service_equipment_names]
        self.service_equipment_names = sorted(set(service_equipment_names))

        if isinstance(turbine_capacities, (float, int)):
            turbine_capacities = [turbine_capacities]
        self.turbine_capacities = turbine_capacities

        if isinstance(events, str):
            events = self._read_data(events)
        self.events = self._apply_inflation_rate(self._tidy_data(events, kind="events"))

        if isinstance(operations, str):
            operations = self._read_data(operations)
        self.operations = self._tidy_data(operations, kind="operations")

        if isinstance(potential, str):
            potential = self._read_data(potential)
        self.potential = self._tidy_data(potential, kind="potential")

        if isinstance(production, str):
            production = self._read_data(production)
        self.production = self._tidy_data(production, kind="production")

        if SAM_settings is not None:
            SAM_settings = "SAM_Singleowner_defaults.yaml"
            try:
                self.sam_settings = load_yaml(
                    self.data_dir / "project/config", SAM_settings
                )
            except FileNotFoundError:
                self.sam_settings = load_yaml(self.data_dir / "windfarm", SAM_settings)
                logging.warning(
                    "DeprecationWarning: In v0.8, all SAM configurations must be"
                    " located in: '<library>/project/config/"
                )

            self._setup_pysam()
        else:
            self.sam_settings = None
            self.financial_model = None

    @classmethod
    def from_simulation_outputs(cls, fpath: Path | str, fname: str) -> Metrics:
        """Creates the Metrics class from the saved outputs of a simulation for ease of
        revisiting the calculated metrics.

        Parameters
        ----------
        fpath : Path | str
            The full path to the file where the data was saved.
        fname : Path | str
            The filename for where the data was saved, which should be a direct
            dictionary mapping for the Metrics initialization.

        Returns
        -------
        Metrics
            The class object.
        """
        data = load_yaml(fpath, fname)
        metrics = cls(**data)
        return metrics

    def _tidy_data(self, data: pd.DataFrame, kind: str) -> pd.DataFrame:
        """Tidies the "raw" csv-converted data to be able to be used among the
        ``Metrics`` class.

        Parameters
        ----------
        data : pd.DataFrame
            The csv log data.
        kind : str
            The category of the input provided to ``data``. Should be one of:
             - "operations"
             - "events"
             - "potential"
             - "production"

        Returns
        -------
        pd.DataFrame
            A tidied data frame to be used for all the operations in this class.
        """
        # Ignore odd pandas casting error for pandas>=1.5(?)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            data = data = data.convert_dtypes()

        if data.index.name != "datetime":
            try:
                data.datetime = pd.to_datetime(data.datetime)
            except AttributeError:
                data["datetime"] = pd.to_datetime(data.env_datetime)
            data.index = data.datetime
            data = data.drop(labels="datetime", axis=1)
        data.env_datetime = pd.to_datetime(data.env_datetime)
        data = data.assign(
            year=data.env_datetime.dt.year,
            month=data.env_datetime.dt.month,
            day=data.env_datetime.dt.day,
        )
        if kind == "operations":
            data[self.turbine_id] = data[self.turbine_id].astype(float)
            turbines = (
                self.turbine_weights[self.turbine_id].values * data[self.turbine_id]
            )
            windfarm = np.sum(
                [
                    data[[sub]]
                    * np.array(
                        [[fsum(row)] for _, row in turbines[val["turbines"]].iterrows()]
                    ).reshape(-1, 1)
                    for sub, val in self.substation_turbine_map.items()
                ],
                axis=0,
            )
            windfarm = pd.DataFrame(windfarm, columns=["windfarm"], index=data.index)
            data = pd.concat([data, windfarm], axis=1)
        return data

    def _read_data(self, fname: str) -> pd.DataFrame:
        """Reads the csv log data from library. This is intended to be used for the
        events or operations data.

        Parameters
        ----------
        path : str
            Path to the simulation library.
        fname : str
            Filename of the csv data.

        Returns
        -------
        pd.DataFrame
            Dataframe of either the events or operations data.
        """
        data = pd.read_csv(
            self.data_dir / "outputs" / "logs" / fname, delimiter="|", engine="pyarrow"
        )
        return data

    def _apply_inflation_rate(self, events: pd.DataFrame) -> pd.DataFrame:
        """Adjusts the cost data for compounding inflation.

        Parameters
        ----------
        inflation_rate : float
            The inflation rate to be applied for each year.
        events : pd.DataFrame
            The events dataframe containing the project cost data.

        Returns
        -------
        pd.DataFrame
            The events dataframe with costs adjusted for inflation.
        """
        adjusted_inflation = deepcopy(self.inflation_rate)
        years = events.year.unique()
        years.sort()
        for year in years:
            row_filter = events.year == year
            if year > years[0]:
                events.loc[row_filter, self._cost_columns] *= adjusted_inflation
                adjusted_inflation *= self.inflation_rate

        return events

    def _setup_pysam(self) -> None:
        """Creates and executes the PySAM model for financial metrics."""
        # Define the model and import the SAM settings file.
        self.financial_model = pysam_singleowner_financial_model.default(
            "WindPowerSingleOwner"
        )
        model_data = pssc.dict_to_ssc_table(self.sam_settings, "singleowner")
        self.financial_model = pysam_singleowner_financial_model.wrap(model_data)

        # Remove the leap year production
        leap_year_ix = self.production.index.month == 2
        leap_year_ix &= self.production.index.day == 29
        generation = self.production.loc[~leap_year_ix].windfarm.values

        # Create a years variable for later use with the PySAM outputs
        self.years = sorted(self.production.year.unique())

        # Let mypy know that I know what I'm doing
        assert isinstance(self.financial_model, PySAM.Singleowner.Singleowner)

        # Replace the coded generation with modeled generation
        self.financial_model.FinancialParameters.analysis_period = len(self.years)
        self.financial_model.SystemOutput.gen = generation

        # Reset the system capacity, in kW
        self.financial_model.FinancialParameters.system_capacity = (
            self.project_capacity * 1000
        )

        # Run the financial model
        self.financial_model.execute()

    def time_based_availability(self, frequency: str, by: str) -> pd.DataFrame:
        """Calculates the time-based availabiliy over a project's lifetime as a single
        value, annual average, or monthly average for the whole windfarm or by turbine.

        .. note:: This currently assumes that if there are multiple substations, that
          the turbines are all connected to multiple.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by : str
            One of "windfarm" or "turbine".

        Returns
        -------
        pd.DataFrame
            The time-based availability at the desired aggregation level.
        """
        frequency = _check_frequency(frequency, which="all")

        by = by.lower().strip()
        if by not in ("windfarm", "turbine"):
            raise ValueError('``by`` must be one of "windfarm" or "turbine".')
        by_turbine = by == "turbine"

        # Determine the operational capacity of each turbine with substation downtime
        operations_cols = ["year", "month", "day", "windfarm"] + self.turbine_id
        turbine_operations = self.operations[operations_cols].copy()
        for sub, val in self.substation_turbine_map.items():
            turbine_operations[val["turbines"]] *= self.operations[[sub]].values

        hourly = turbine_operations.loc[:, self.turbine_id].values

        # TODO: The below should be better summarized as:
        # (availability > 0).groupby().sum() / groupby().count()

        if frequency == "project":
            availability = _calculate_time_availability(hourly, by_turbine=by_turbine)
            if not by_turbine:
                return pd.DataFrame([availability], columns=["windfarm"])
            availability = pd.DataFrame(
                availability.reshape(1, -1), columns=self.turbine_id  # type: ignore
            )
            return availability
        elif frequency == "annual":
            date_time = turbine_operations[["year"]]
            counts = turbine_operations.groupby(by="year").count()
            counts = counts[self.turbine_id] if by_turbine else counts[["windfarm"]]
            annual = [
                _calculate_time_availability(
                    hourly[date_time.year == year], by_turbine=by_turbine
                )
                for year in counts.index
            ]
            return pd.DataFrame(annual, index=counts.index, columns=counts.columns)
        elif frequency == "monthly":
            date_time = turbine_operations[["month"]]
            counts = turbine_operations.groupby(by="month").count()
            counts = counts[self.turbine_id] if by_turbine else counts[["windfarm"]]
            monthly = [
                _calculate_time_availability(
                    hourly[date_time.month == month], by_turbine=by_turbine
                )
                for month in counts.index
            ]
            return pd.DataFrame(monthly, index=counts.index, columns=counts.columns)
        elif frequency == "month-year":
            date_time = turbine_operations[["year", "month"]]
            counts = turbine_operations.groupby(by=["year", "month"]).count()
            counts = counts[self.turbine_id] if by_turbine else counts[["windfarm"]]
            month_year = [
                _calculate_time_availability(
                    hourly[(date_time.year == year) & (date_time.month == month)],
                    by_turbine=by_turbine,
                )
                for year, month in counts.index
            ]
            return pd.DataFrame(month_year, index=counts.index, columns=counts.columns)

    def production_based_availability(self, frequency: str, by: str) -> pd.DataFrame:
        """Calculates the production-based availabiliy over a project's lifetime as a
        single value, annual average, or monthly average for the whole windfarm or by
        turbine.

        .. note:: This currently assumes that if there are multiple substations, that
          the turbines are all connected to multiple.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by : str
            One of "windfarm" or "turbine".

        Returns
        -------
        pd.DataFrame
            The production-based availability at the desired aggregation level.
        """
        frequency = _check_frequency(frequency, which="all")

        by = by.lower().strip()
        if by not in ("windfarm", "turbine"):
            raise ValueError('``by`` must be one of "windfarm" or "turbine".')
        by_turbine = by == "turbine"

        production = self.production.loc[:, self.turbine_id]
        potential = self.potential.loc[:, self.turbine_id]

        if frequency == "project":
            production = production.values
            potential = potential.values
            if (potential == 0).sum() > 0:
                potential[potential == 0] = 1
            if not by_turbine:
                return pd.DataFrame(
                    [production.sum() / potential.sum()], columns=["windfarm"]
                )
            availability = pd.DataFrame(
                (production.sum(axis=0) / potential.sum(axis=0)).reshape(1, -1),
                columns=self.turbine_id,
            )
            return availability

        production["year"] = production.index.year.values
        production["month"] = production.index.month.values

        potential["year"] = potential.index.year.values
        potential["month"] = potential.index.month.values

        group_cols = deepcopy(self.turbine_id)
        if frequency == "annual":
            group_cols.insert(0, "year")
            production = production[group_cols].groupby("year").sum()
            potential = potential[group_cols].groupby("year").sum()

        elif frequency == "monthly":
            group_cols.insert(0, "month")
            production = production[group_cols].groupby("month").sum()
            potential = potential[group_cols].groupby("month").sum()

        elif frequency == "month-year":
            group_cols.insert(0, "year")
            group_cols.insert(0, "month")
            production = production[group_cols].groupby(["year", "month"]).sum()
            potential = potential[group_cols].groupby(["year", "month"]).sum()

        if (potential.values == 0).sum() > 0:
            potential.loc[potential.values == 0] = 1
        columns = self.turbine_id
        if not by_turbine:
            production = production.sum(axis=1)
            potential = potential.sum(axis=1)
            columns = [by]
        return pd.DataFrame(production / potential, columns=columns)

    def capacity_factor(  # type: ignore
        self, which: str, frequency: str, by: str
    ) -> pd.DataFrame:
        """Calculates the capacity factor over a project's lifetime as a single value,
        annual average, or monthly average for the whole windfarm or by turbine.

        .. note:: This currently assumes that if there are multiple substations, that
          the turbines are all connected to multiple.


        Parameters
        ----------
        which : str
            One of "net" or "gross".
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by : str
            One of "windfarm" or "turbine".

        Returns
        -------
        pd.DataFrame
            The capacity factor at the desired aggregation level.
        """
        which = which.lower().strip()
        if which not in ("net", "gross"):
            raise ValueError('``which`` must be one of "net" or "gross".')

        frequency = _check_frequency(frequency, which="all")

        by = by.lower().strip()
        if by not in ("windfarm", "turbine"):
            raise ValueError('``by`` must be one of "windfarm" or "turbine".')
        by_turbine = by == "turbine"

        production = self.production if which == "net" else self.potential
        production = production.loc[:, self.turbine_id]

        if frequency == "project":
            if not by_turbine:
                potential = production.shape[0] * self.project_capacity * 1000.0
                production = production.values.sum()
                return pd.DataFrame([production / potential], columns=["windfarm"])

            potential = production.shape[0] * np.array(self.turbine_capacities)
            return pd.DataFrame(production.sum(axis=0) / potential).T

        production["year"] = production.index.year.values
        production["month"] = production.index.month.values

        if frequency == "annual":
            group_cols = ["year"]
        elif frequency == "monthly":
            group_cols = ["month"]
        elif frequency == "month-year":
            group_cols = ["year", "month"]

        potential = production[group_cols + self.turbine_id].groupby(group_cols).count()
        production = production[group_cols + self.turbine_id].groupby(group_cols).sum()

        if by_turbine:
            capacity = np.array(self.turbine_capacities, dtype=float)
            columns = self.turbine_id
            potential *= capacity
        else:
            capacity = self.project_capacity
            production = production.sum(axis=1)
            columns = [by]
        return pd.DataFrame(production / potential, columns=columns)

    def task_completion_rate(self, which: str, frequency: str) -> float | pd.DataFrame:
        """Calculates the task completion rate over a project's lifetime as a single
        value, annual average, or monthly average for the whole windfarm or by turbine.

        .. note:: This currently assumes that if there are multiple substations, that
          the turbines are all connected to multiple.

        Parameters
        ----------
        which : str
            One of "scheduled", "unscheduled", or "both".
        frequency : str
            One of "project", "annual", "monthly", or "month-year".

        Returns
        -------
        float | pd.DataFrame
            The task completion rate at the desired aggregation level.
        """
        which = which.lower().strip()
        if which not in ("scheduled", "unscheduled", "both"):
            raise ValueError(
                '``which`` must be one of "scheduled", "unscheduled", or "both".'
            )

        frequency = _check_frequency(frequency, which="all")

        if which == "scheduled":
            task_filter = ["maintenance"]
        elif which == "unscheduled":
            task_filter = ["repair"]
        else:
            task_filter = ["maintenance", "repair"]

        cols = ["env_datetime", "request_id"]
        completion_filter = [f"{el} complete" for el in task_filter]
        request_filter = [f"{el} request" for el in task_filter]
        requests = self.events.loc[
            self.events.action.isin(request_filter), cols
        ].reset_index(drop=True)
        completions = self.events.loc[
            self.events.action.isin(completion_filter), cols
        ].reset_index(drop=True)

        if frequency == "project":
            if requests.shape[0] == 0:
                return pd.DataFrame([0.0], columns=["windfarm"])
            return pd.DataFrame(
                [completions.shape[0] / requests.shape[0]], columns=["windfarm"]
            )

        requests["year"] = requests.env_datetime.dt.year.values
        requests["month"] = requests.env_datetime.dt.month.values

        completions["year"] = completions.env_datetime.dt.year.values
        completions["month"] = completions.env_datetime.dt.month.values

        if frequency == "annual":
            group_filter = ["year"]
            indices = self.operations.year.unique()
        elif frequency == "monthly":
            group_filter = ["month"]
            indices = self.operations.month.unique()
        elif frequency == "month-year":
            group_filter = ["year", "month"]
            indices = (
                self.operations[["year", "month"]]
                .groupby(["year", "month"])
                .value_counts()
                .index.tolist()
            )

        group_cols = group_filter + ["request_id"]
        requests = requests[group_cols].groupby(group_filter).count()
        requests.loc[requests.request_id == 0] = 1
        completions = completions[group_cols].groupby(group_filter).count()

        missing = [ix for ix in indices if ix not in requests.index]
        requests = pd.concat(
            [
                requests,
                pd.DataFrame(
                    np.ones(len(missing)), index=missing, columns=requests.columns
                ),
            ]
        ).sort_index()

        missing = [ix for ix in indices if ix not in completions.index]
        completions = pd.concat(
            [
                completions,
                pd.DataFrame(
                    np.ones(len(missing)), index=missing, columns=completions.columns
                ),
            ]
        ).sort_index()

        completion_rate = pd.DataFrame(completions / requests)
        completion_rate.index = completion_rate.index.set_names(group_filter)
        return completion_rate.rename(
            columns={"request_id": "Completion Rate", 0: "Completion Rate"}
        )

    def equipment_costs(
        self, frequency: str, by_equipment: bool = False
    ) -> pd.DataFrame:
        """Calculates the equipment costs for the simulation at a project, annual, or
        monthly level with (or without) respect to equipment utilized in the simulation.
        This excludes any port fees that might apply, which are included in:
        ``port_fees``.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_equipment : bool, optional
            Indicates whether the values are with resepect to the equipment utilized
            (True) or not (False), by default False.

        Returns
        -------
        pd.DataFrame
            Returns pandas ``DataFrame`` with columns:
                - year (if appropriate for frequency)
                - month (if appropriate for frequency)
                - then any equipment names as they appear in the logs

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_equipment`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")

        if not isinstance(by_equipment, bool):
            raise ValueError("`by_equipment` must be one of `True` or `False`")

        if frequency == "annual":
            col_filter = ["year"]
        elif frequency == "monthly":
            col_filter = ["month"]
        elif frequency == "month-year":
            col_filter = ["year", "month"]

        cost_col = [self._equipment_cost]
        events = self.events.loc[self.events.action != "monthly lease fee"]
        if by_equipment:
            if frequency == "project":
                costs = (
                    events.loc[events[self._equipment_cost] > 0, cost_col + ["agent"]]
                    .groupby(["agent"])
                    .sum()
                    .fillna(0)
                    .reset_index(level=0)
                )
                costs = costs.fillna(costs.max(axis=0)).T
                costs = costs.rename(columns=costs.iloc[0]).drop(index="agent")
                return costs.reset_index(drop=True)

            col_filter = ["agent"] + col_filter
            costs = (
                events.loc[events[self._equipment_cost] > 0, cost_col + col_filter]
                .groupby(col_filter)
                .sum()
                .reset_index(level=0)
            )
            costs = pd.concat(
                [
                    costs[costs.agent == eq][cost_col].rename(
                        columns={self._equipment_cost: eq}
                    )
                    for eq in costs.agent.unique()
                ],
                axis=1,
            )
            return costs.fillna(value=0)

        if frequency == "project":
            return pd.DataFrame([events[cost_col].sum()], columns=cost_col)

        costs = events[cost_col + col_filter].groupby(col_filter).sum()
        return costs.fillna(0)

    def service_equipment_utilization(self, frequency: str) -> pd.DataFrame:
        """Calculates the utilization rate for each of the service equipment in the
        simulation  as the ratio of total number of days each of the servicing
        equipment is in operation over the total number of days it's present in the
        simulation. This number excludes mobilization time and the time between
        visits for scheduled servicing equipment strategies.

        .. note:: For tugboats in a tow-to-port scenario, this ratio will be near
            100% because they are considered to be operating on an as-needed basis per
            the port contracting assumptions

        Parameters
        ----------
        frequency : str
            One of "project" or "annual".

        Returns
        -------
        pd.DataFrame
            The utilization rate of each of the simulation ``SerivceEquipment``.

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project" or "annual".
        """
        frequency = _check_frequency(frequency, which="annual")

        operation_days = []
        total_days = []
        operating_actions = [
            "traveling",  # traveling between port/site or on-site
            "repair",
            "maintenance",
            "delay",  # performing work
            "unmooring",
            "mooring_reconnection",
            "towing",  # tugboat classifications
        ]
        operating_filter = self.events.action.isin(operating_actions)
        return_filter = self.events.action == "delay"
        return_filter &= (
            (self.events.reason == "work is complete")
            & (self.events.additional == "will return next year")
        ) | (self.events.reason == "non-operational period")
        return_filter &= self.events.additional == "will return next year"
        for name in self.service_equipment_names:
            equipment_filter = self.events.agent == name
            _events = self.events[equipment_filter & operating_filter]
            _events = _events.groupby(["year", "month", "day"]).size()
            _events = _events.reset_index().groupby("year").count()[["day"]]
            operation_days.append(_events.rename(columns={"day": name}))

            ix_filter = equipment_filter & ~return_filter
            total = self.events[ix_filter].groupby(["year", "month", "day"]).size()
            total = total.reset_index().groupby("year").count()[["day"]]
            total_days.append(total.rename(columns={"day": name}))

        operating_df = pd.DataFrame(operation_days[0])
        total_df = pd.DataFrame(total_days[0])
        if len(self.service_equipment_names) > 1:
            operating_df = operating_df.join(operation_days[1:], how="outer").fillna(0)
            total_df = total_df.join(total_days[1:], how="outer").fillna(1)

        for year in self.events.year.unique():
            if year not in operating_df.index:
                missing = pd.DataFrame(
                    np.zeros((1, operating_df.shape[1])),
                    index=[year],
                    columns=operating_df.columns,
                )
                operating_df = operating_df.append(missing).sort_index()
            if year not in total_df.index:
                missing = pd.DataFrame(
                    np.ones((1, total_df.shape[1])),
                    index=[year],
                    columns=operating_df.columns,
                )
                total_df = total_df.append(missing).sort_index()

        if frequency == "project":
            operating_df = operating_df.reset_index().sum()[
                self.service_equipment_names
            ]
            total_df = total_df.reset_index().sum()[self.service_equipment_names]
            return pd.DataFrame(operating_df / total_df).T
        return operating_df / total_df

    def vessel_crew_hours_at_sea(
        self,
        frequency: str,
        by_equipment: bool = False,
        vessel_crew_assumption: dict[str, float] = {},
    ) -> pd.DataFrame:
        """Calculates the total number of crew hours at sea that occurred during a
        simulation at a project, annual, or monthly level that can be broken out by
        servicing equipment. This includes time mobilizing, delayed at sea, servicing,
        towing, and traveling.

        .. note:: This metric is intended to be used for offshore wind simulations.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_equipment : bool, optional
            Indicates whether the values are with resepect to each tugboat (True) or not
            (False), by default False.
        vessel_crew_assumption : dict[str, float], optional
            Dictionary of vessel names (``ServiceEquipment.settings.name``) and number
            of crew members aboard to trannsform the results from vessel hours at sea
            to crew hours at sea.

        Returns
        -------
        pd.DataFrame
            Returns a pandas ``DataFrame`` with columns:

            - year (if appropriate for frequency)
            - month (if appropriate for frequency)
            - Total Crew Hours at Sea
            - {ServiceEquipment.settings.name} (if broken out)

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_equipment`` is not one of ``True`` or ``False``.
        ValueError
            If ``vessel_crew_assumption`` is not a dictionary.
        """
        frequency = _check_frequency(frequency, which="all")

        if not isinstance(by_equipment, bool):
            raise ValueError("``by_equipment`` must be one of ``True`` or ``False``")

        if not isinstance(vessel_crew_assumption, dict):
            raise ValueError(
                "`vessel_crew_assumption` must be a dictionary of vessel name (keys)"
                " and number of crew (values)"
            )

        # Filter by the at sea indicators and required columns
        at_sea = self.events
        at_sea = at_sea.loc[
            at_sea.location.isin(("enroute", "site", "system"))
            & at_sea.agent.isin(self.service_equipment_names),
            ["agent", "year", "month", "action", "reason", "additional", "duration"],
        ].reset_index(drop=True)

        # Create a shell for the final results
        total_hours = (
            self.events[["env_time", "year", "month"]]
            .groupby(["year", "month"])
            .count()
        )
        total_hours = total_hours.reset_index().rename(columns={"env_time": "N"})
        total_hours.N = 0

        # Apply the vessel crew assumptions
        vessels = at_sea.agent.unique()
        if vessel_crew_assumption != {}:
            for name, n_crew in vessel_crew_assumption.items():
                if name not in vessels:
                    continue
                ix_vessel = at_sea.agent == name
                at_sea.loc[ix_vessel, "duration"] *= n_crew

        group_cols = ["agent"]
        columns = ["Total Crew Hours at Sea"] + vessels.tolist()
        if not by_equipment:
            group_cols.pop(0)
            columns = ["Total Crew Hours at Sea"]
            at_sea = at_sea.groupby(["year", "month"]).sum()[["duration"]].reset_index()

        if frequency == "project":
            total_hours = pd.DataFrame([[0]], columns=["duration"])
            if by_equipment:
                total_hours = (
                    at_sea[["duration", "agent"]]
                    .groupby(["agent"])
                    .sum()
                    .T.reset_index(drop=True)
                )
                total_hours.loc[:, "Total Crew Hours at Sea"] = total_hours.sum().sum()
                return total_hours[columns]
            else:
                return pd.DataFrame(at_sea.sum()[["duration"]]).T.rename(
                    columns={"duration": "Total Crew Hours at Sea"}
                )

        elif frequency == "annual":
            additional_cols = ["year"]
            total_hours = total_hours.groupby("year")[["N"]].sum()
        elif frequency == "monthly":
            additional_cols = ["month"]
            total_hours = total_hours.groupby("month")[["N"]].sum()
        elif frequency == "month-year":
            additional_cols = ["year", "month"]
            total_hours = total_hours.groupby(["year", "month"])[["N"]].sum()

        columns = additional_cols + columns
        group_cols.extend(additional_cols)
        at_sea = at_sea[group_cols + ["duration"]].groupby(group_cols).sum()
        if by_equipment:
            total = []
            for v in vessels:
                total.append(at_sea.loc[v].rename(columns={"duration": v}))
            total_hours = total_hours.join(
                pd.concat(total, axis=1), how="outer"
            ).fillna(0)
            total_hours.N = total_hours.sum(axis=1)
            total_hours = (
                total_hours.reset_index()
                .rename(columns={"N": "Total Crew Hours at Sea"})[columns]
                .set_index(additional_cols)
            )
            return total_hours

        return at_sea.rename(columns={"duration": "Total Crew Hours at Sea"})

    def number_of_tows(
        self, frequency: str, by_tug: bool = False, by_direction: bool = False
    ) -> float | pd.DataFrame:
        """Calculates the total number of tows that occurred during a simulation at a
        project, annual, or monthly level that can be broken out by tugboat.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_tug : bool, optional
            Indicates whether the values are with resepect to each tugboat (True) or not
            (False), by default False.
        by_direction : bool, optional
            Indicates whether the values are with respect to the direction a turbine is
            towed (True) or not (False), by default False.

        Returns
        -------
        float | pd.DataFrame
            Returns either a float for whole project-level costs or a pandas
            ``DataFrame`` with columns:

            - year (if appropriate for frequency)
            - month (if appropriate for frequency)
            - total_tows
            - total_tows_to_port (if broken out)
            - total_tows_to_site (if broken out)
            - {ServiceEquipment.settings.name}_total_tows (if broken out)
            - {ServiceEquipment.settings.name}_to_port (if broken out)
            - {ServiceEquipment.settings.name}_to_site (if broken out)

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_tug`` is not one of ``True`` or ``False``.
        ValueError
            If ``by_direction`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")

        if not isinstance(by_tug, bool):
            raise ValueError("``by_tug`` must be one of ``True`` or ``False``")

        if not isinstance(by_direction, bool):
            raise ValueError("``by_direction`` must be one of ``True`` or ``False``")

        # Filter out only the towing events
        towing = self.events.loc[self.events.action == "towing"].copy()
        if towing.shape[0] == 0:
            # If this is accessed in an in-situ only scenario, or no tows were activated
            # then return back 0
            return pd.DataFrame([[0]], columns=["total_tows"])
        towing.loc[:, "direction"] = "to_site"
        ix_to_port = towing.reason == "towing turbine to port"
        towing.loc[ix_to_port, "direction"] = "to_port"

        # Get the unique directions and tugboat names
        direction_suffix = ("to_port", "to_site")
        tugboats = towing.agent.unique().tolist()

        # Create the final column names
        columns = ["total_tows"]
        if by_direction:
            columns.extend([f"{c}_{s}" for c in columns for s in direction_suffix])

        if by_tug:
            tug_columns = [f"{t}_total_tows" for t in tugboats]
            if by_direction:
                _columns = [f"{t}_{s}" for t in tugboats for s in direction_suffix]
                tug_columns.extend(_columns)
                tug_columns.sort()
            columns.extend(tug_columns)

        # Count the total number of tows by each possibly category
        n_tows = towing.groupby(["agent", "year", "month", "direction"]).count()
        n_tows = n_tows.rename(columns={"env_time": "N"})["N"].reset_index()

        # Create a shell for the total tows
        total_tows = (
            self.events[["env_time", "year", "month"]]
            .groupby(["year", "month"])
            .count()["env_time"]
        )
        total_tows = total_tows.reset_index().rename(columns={"env_time": "N"})
        total_tows.N = 0

        # Create the correct time frequency for the number of tows and shell total
        group_cols = ["agent", "direction"]
        if frequency == "project":
            time_cols = []
            n_tows = n_tows[group_cols + ["N"]].groupby(group_cols).sum()

            # If no further work is required, then return the sum as a 1x1 data frame
            if not by_tug and not by_direction:
                return pd.DataFrame(
                    [n_tows.reset_index().N.sum()], columns=["total_tows"]
                )

            total_tows = pd.DataFrame([[0]], columns=["N"])
        elif frequency == "annual":
            time_cols = ["year"]
            columns = time_cols + columns
            group_cols.extend(time_cols)
            n_tows = n_tows.groupby(group_cols).sum()[["N"]]
            total_tows = (
                total_tows[["year", "N"]].groupby(time_cols).sum().reset_index()
            )
        elif frequency == "monthly":
            time_cols = ["month"]
            columns = time_cols + columns
            group_cols.extend(time_cols)
            n_tows = n_tows[group_cols + ["N"]].groupby(group_cols).sum()
            total_tows = (
                total_tows[["month", "N"]].groupby(time_cols).sum().reset_index()
            )
        elif frequency == "month-year":
            # Already have month-year by default, so skip the n_tows refinement
            time_cols = ["year", "month"]
            columns = time_cols + columns
            group_cols.extend(time_cols)
            n_tows = n_tows.set_index(group_cols, drop=True)

        # Create a list of the columns needed for creating the broken down totals
        if frequency == "project":
            total_cols = ["N"]
        else:
            total_cols = total_tows.drop(columns=["N"]).columns.tolist()

        # Sum the number of tows by tugboat, if needed
        if by_tug:
            tug_sums = []
            for tug in tugboats:
                tug_sum = n_tows.loc[tug]
                tug_sums.append(tug_sum.rename(columns={"N": tug}))
            tug_sums_by_direction = pd.concat(tug_sums, axis=1).fillna(0)

            if frequency == "project":
                tug_sums = pd.DataFrame(tug_sums_by_direction.sum()).T
            else:
                tug_sums = tug_sums_by_direction.reset_index().groupby(total_cols).sum()
            assert isinstance(tug_sums, pd.DataFrame)  # mypy checking
            tug_sums = tug_sums.rename(
                columns={t: f"{t}_total_tows" for t in tug_sums.columns}
            )
            assert isinstance(tug_sums, pd.DataFrame)  # mypy checking
            total = pd.DataFrame(
                tug_sums.sum(axis=1), columns=["total_tows"]
            ).reset_index()
        else:
            if not by_direction:
                # Sum the totals, then merge the results with the shell data frame,
                # and cleanup the columns
                total = n_tows.reset_index().groupby(total_cols).sum().reset_index()
                total_tows = total_tows.merge(total, on=total_cols, how="outer")
                total_tows = total_tows.fillna(0).rename(columns={"N_y": "total_tows"})
                total_tows = total_tows[columns]
                if time_cols:
                    return total_tows.set_index(time_cols)
                return total_tows
            else:
                total = (
                    n_tows.groupby(total_cols)
                    .sum()
                    .reset_index()
                    .rename(columns={"N": "total_tows"})
                )

        # Create the full total tows data
        if frequency == "project":
            if "index" in total.columns:
                total_tows = total.drop(columns=["index"])
        else:
            total_tows = (
                total_tows.merge(total, how="outer").drop(columns=["N"]).fillna(0)
            )
            total_tows = total_tows.set_index(total_cols)

        # Get the sums by each direction towed, if needed
        if by_direction:
            if frequency == "project":
                direction_sums = n_tows.reset_index().groupby("direction").sum()
                for s in direction_suffix:
                    total_tows.loc[:, f"total_tows_{s}"] = direction_sums.loc[s, "N"]
            else:
                direction_sums = (
                    n_tows.reset_index().groupby(["direction"] + total_cols).sum()
                )
                for s in direction_suffix:
                    total_tows = total_tows.join(
                        direction_sums.loc[s].rename(columns={"N": f"total_tows_{s}"})
                    ).fillna(0)

            # Add in the tugboat breakdown as needed
            if by_tug:
                total_tows = total_tows.join(tug_sums, how="outer").fillna(0)
                for s in direction_suffix:
                    if frequency == "project":
                        _total = pd.DataFrame(
                            tug_sums_by_direction.loc[s]
                        ).T.reset_index(drop=True)
                    else:
                        _total = tug_sums_by_direction.loc[s]
                    total_tows = total_tows.join(
                        _total.rename(columns={t: f"{t}_{s}" for t in tugboats}),
                        how="outer",
                    ).fillna(0)
                total_tows = total_tows.reset_index()[columns]
                if time_cols:
                    return total_tows.set_index(time_cols)
                else:
                    return total_tows
            total_tows = total_tows.assign(N=total_tows.sum(axis=1))
            total_tows = total_tows.rename(columns={"N": "total_tows"}).reset_index()[
                columns
            ]
            if time_cols:
                return total_tows.set_index(time_cols)
            return total_tows

        if by_tug:
            total_tows = (
                total_tows.join(tug_sums, how="outer").fillna(0).reset_index()[columns]
            )
            if time_cols:
                return total_tows.set_index(time_cols)
            return total_tows

        if time_cols:
            return total_tows[columns].set_index(time_cols)
        return total_tows[columns]

    def labor_costs(
        self, frequency: str, by_type: bool = False
    ) -> float | pd.DataFrame:
        """Calculates the labor costs for the simulation at a project, annual, or
        monthly level that can be broken out by hourly and salary labor costs.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_type : bool, optional
            Indicates whether the values are with resepect to the labor types
            (True) or not (False), by default False.

        Returns
        -------
        float | pd.DataFrame
            Returns either a float for whole project-level costs or a pandas
            ``DataFrame`` with columns:

            - year (if appropriate for frequency)
            - month (if appropriate for frequency)
            - total_labor_cost
            - hourly_labor_cost (if broken out)
            - salary_labor_cost (if broken out)

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_type`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")

        if not isinstance(by_type, bool):
            raise ValueError("``by_type`` must be one of ``True`` or ``False``")

        labor_cols = [self._hourly_cost, self._salary_cost, self._labor_cost]
        if frequency == "project":
            costs = pd.DataFrame(
                self.events[labor_cols].sum(axis=0).values.reshape(1, -1),
                columns=labor_cols,
            )
            if not by_type:
                return costs[[self._labor_cost]]
            return costs

        if frequency == "annual":
            group_filter = ["year"]
        elif frequency == "monthly":
            group_filter = ["month"]
        elif frequency == "month-year":
            group_filter = ["year", "month"]

        costs = (
            self.events.loc[:, labor_cols + group_filter]
            .groupby(group_filter)
            .sum()
            .fillna(value=0)
        )
        if not by_type:
            return pd.DataFrame(costs[self._labor_cost])
        return costs

    def equipment_labor_cost_breakdowns(
        self, frequency: str, by_category: bool = False
    ) -> pd.DataFrame:
        """Calculates the producitivty cost breakdowns for the simulation at a project,
        annual, or monthly level that can be broken out to include the equipment and
        labor components.

        .. note:: Doesn't produce a value if there's no cost associated with a "reason".

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_category : bool, optional
            Indicates whether to include the equipment and labor categories (True) or
            not (False), by default False.

        Returns
        -------
        pd.DataFrame
            Returns pandas ``DataFrame`` with columns:
                - year (if appropriate for frequency)
                - month (if appropriate for frequency)
                - reason
                - hourly_labor_cost (if by_category == ``True``)
                - salary_labor_cost (if by_category == ``True``)
                - total_labor_cost (if by_category == ``True``)
                - equipment_cost (if by_category == ``True``)
                - total_cost (if broken out)

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_category`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")
        if not isinstance(by_category, bool):
            raise ValueError("``by_equipment`` must be one of ``True`` or ``False``")

        group_filter = ["action", "reason", "additional"]
        if frequency in ("annual", "month-year"):
            group_filter.insert(0, "year")
        elif frequency == "monthly":
            group_filter.insert(0, "month")
        if frequency == "month-year":
            group_filter.insert(1, "month")

        action_list = [
            "delay",
            "repair",
            "maintenance",
            "mobilization",
            "transferring crew",
            "traveling",
        ]
        equipment = self.events[self.events[self._equipment_cost] > 0].agent.unique()
        costs = (
            self.events.loc[
                self.events.agent.isin(equipment)
                & self.events.action.isin(action_list)
                & ~self.events.additional.isin(["work is complete"]),
                group_filter + self._cost_columns,
            ]
            .groupby(group_filter)
            .sum()
            .reset_index()
        )
        costs["display_reason"] = [""] * costs.shape[0]

        non_shift_hours = (
            "not in working hours",
            "work shift has ended; waiting for next shift to start",
            "no more return visits will be made",
            "will return next year",
            "waiting for next operational period",
        )
        weather_hours = ("weather delay", "weather unsuitable to transfer crew")
        costs.loc[
            (costs.action == "delay") & (costs.additional.isin(non_shift_hours)),
            "display_reason",
        ] = "Not in Shift"
        costs.loc[costs.action == "repair", "display_reason"] = "Repair"
        costs.loc[costs.action == "maintenance", "display_reason"] = "Maintenance"
        costs.loc[
            costs.action == "transferring crew", "display_reason"
        ] = "Crew Transfer"
        costs.loc[costs.action == "traveling", "display_reason"] = "Site Travel"
        costs.loc[costs.action == "mobilization", "display_reason"] = "Mobilization"
        costs.loc[
            costs.additional.isin(weather_hours), "display_reason"
        ] = "Weather Delay"
        costs.loc[costs.reason == "no requests", "display_reason"] = "No Requests"

        costs.reason = costs.display_reason

        drop_columns = [self._materials_cost, "display_reason", "additional", "action"]
        if not by_category:
            drop_columns.extend(
                [
                    self._hourly_cost,
                    self._salary_cost,
                    self._labor_cost,
                    self._equipment_cost,
                ]
            )
        group_filter.pop(group_filter.index("additional"))
        group_filter.pop(group_filter.index("action"))
        costs = costs.drop(columns=drop_columns)
        costs = costs.groupby(group_filter).sum().reset_index()

        month_year = frequency == "month-year"
        if frequency in ("annual", "month-year"):
            years = costs.year.unique()
            reasons = costs.reason.unique()
            comparison_values = product(years, reasons)
            if month_year:
                months = costs.month.unique()
                comparison_values = product(years, months, reasons)

            zeros = np.zeros(costs.shape[1] - 2).tolist()
            for _year, *_month, _reason in comparison_values:
                row_filter = costs.year.values == _year
                row = [_year, _reason] + zeros
                if month_year:
                    _month = _month[0]
                    row_filter &= costs.month.values == _month
                    row = [_year, _month, _reason] + zeros[:-1]

                row_filter &= costs.reason.values == _reason
                if costs.loc[row_filter].size > 0:
                    continue
                costs.loc[costs.shape[0]] = row
        elif frequency == "monthly":
            months = costs.month.unique()
            reasons = costs.reason.unique()
            comparison_values = product(months, reasons)
            zeros = np.zeros(costs.shape[1] - 2).tolist()
            for _month, _reason in comparison_values:
                row_filter = costs.month.values == _month
                row_filter &= costs.reason.values == _reason
                row = [_month, _reason] + zeros
                if costs.loc[row_filter].size > 0:
                    continue
                costs.loc[costs.shape[0]] = row

        new_sort = [
            "Maintenance",
            "Repair",
            "Crew Transfer",
            "Site Travel",
            "Mobilization",
            "Weather Delay",
            "No Requests",
            "Not in Shift",
        ]
        costs.reason = pd.Categorical(costs.reason, new_sort)
        costs = costs.set_index(group_filter)
        if frequency == "project":
            return costs.sort_values(by="reason")
        if frequency == "annual":
            return costs.sort_values(by=["year", "reason"])
        if frequency == "monthly":
            return costs.sort_values(by=["month", "reason"])
        return costs.sort_values(by=["year", "month", "reason"])

    def component_costs(
        self, frequency: str, by_category: bool = False, by_action: bool = False
    ) -> pd.DataFrame:
        """Calculates the component costs for the simulation at a project, annual, or
        monthly level that can be broken out by cost categories. This will not sum to
        the total cost because it is does not include times where there is no work being
        done, but costs are being accrued.

        .. note:: It should be noted that the costs will include costs accrued from both
           weather delays and shift-to-shift delays. In the future these will be
           disentangled.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by_category : bool, optional
            Indicates whether the values are with resepect to the various cost
            categories (True) or not (False), by default False.
        by_action : bool, optional
            Indicates whether component costs are going to be further broken out by the
            action being performed--repair, maintenance, and delay--(True) or not
            (False), by default False.

        Returns
        -------
        float | pd.DataFrame
            Returns either a float for whole project-level costs or a pandas
            ``DataFrame`` with columns:

            - year (if appropriate for frequency)
            - month (if appropriate for frequency)
            - component
            - action (if broken out)
            - materials_cost (if broken out)
            - total_labor_cost (if broken out)
            - equipment_cost (if broken out)
            - total_cost

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_category`` is not one of ``True`` or ``False``.
        ValueError
            If ``by_action`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")
        if not isinstance(by_category, bool):
            raise ValueError("``by_equipment`` must be one of ``True`` or ``False``")
        if not isinstance(by_action, bool):
            raise ValueError("``by_equipment`` must be one of ``True`` or ``False``")

        part_filter = ~self.events.part_id.isna() & ~self.events.part_id.isin([""])
        events = self.events.loc[part_filter].copy()

        # Need to simplify the cable identifiers to exclude the connection information
        events.loc[:, "component"] = [el.split("::")[0] for el in events.part_id.values]

        group_filter = []
        if frequency == "annual":
            group_filter.extend(["year"])
        elif frequency == "monthly":
            group_filter.extend(["month"])
        elif frequency == "month-year":
            group_filter.extend(["year", "month"])

        group_filter.append("component")
        cost_cols = ["total_cost"]
        if by_category:
            cost_cols[0:0] = [
                self._materials_cost,
                self._labor_cost,
                self._equipment_cost,
            ]

        if by_action:
            repair_map = {
                val: "repair" for val in ("repair request", "repair", "repair_complete")
            }
            maintenance_map = {
                val: "maintenance"
                for val in (
                    "maintenance request",
                    "maintenance",
                    "maintenance_complete",
                )
            }
            delay_map = {"delay": "delay"}
            action_map = {**repair_map, **maintenance_map, **delay_map}
            events.action = events.action.map(action_map)
            group_filter.append("action")

        month_year = frequency == "month-year"
        zeros = np.zeros(len(cost_cols)).tolist()
        costs = (
            events[group_filter + cost_cols].groupby(group_filter).sum().reset_index()
        )
        if not by_action:
            costs.loc[:, "action"] = np.zeros(costs.shape[0])
            cols = costs.columns.to_list()
            _ix = cols.index("component") + 1
            cols[_ix:_ix] = ["action"]
            cols.pop(-1)
            costs = costs.loc[:, cols]
        if frequency in ("annual", "month-year"):
            years = costs.year.unique()
            components = costs.component.unique()
            actions = costs.action.unique()
            comparison_values = product(years, components, actions)
            if month_year:
                months = costs.month.unique()
                comparison_values = product(years, months, components, actions)

            for _year, *_month, _component, _action in comparison_values:
                row_filter = costs.year.values == _year
                row_filter &= costs.component.values == _component
                row_filter &= costs.action.values == _action
                row = [_year, _component, _action] + zeros
                if month_year:
                    _month = _month[0]
                    row_filter &= costs.month.values == _month
                    row = [_year, _month, _component, _action] + zeros

                if costs.loc[row_filter].size > 0:
                    continue
                costs.loc[costs.shape[0]] = row
        elif frequency == "monthly":
            months = costs.month.unique()
            components = costs.component.unique()
            actions = costs.action.unique()
            comparison_values = product(months, actions, components)
            for _month, _action, _component in comparison_values:
                row_filter = costs.month.values == _month
                row_filter &= costs.component.values == _component
                row_filter &= costs.action.values == _action
                row = [_month, _component, _action] + zeros
                if costs.loc[row_filter].size > 0:
                    continue
                costs.loc[costs.shape[0]] = row
        elif frequency == "project":
            components = costs.component.unique()
            actions = costs.action.unique()
            comparison_values = product(actions, components)
            for _action, _component in comparison_values:
                row_filter = costs.component.values == _component
                row_filter &= costs.action.values == _action
                row = [_component, _action] + zeros
                if costs.loc[row_filter].size > 0:
                    continue
                costs.loc[costs.shape[0]] = row
        sort_cols = group_filter + cost_cols
        if group_filter != []:
            costs = costs.sort_values(group_filter)
        if sort_cols != []:
            costs = costs.loc[:, sort_cols]
        costs = costs.reset_index(drop=True)
        return costs if group_filter == [] else costs.set_index(group_filter)

    def port_fees(self, frequency: str) -> pd.DataFrame:
        """Calculates the port fees for the simulation at a project, annual, or monthly
        level. This excludes any equipment or labor costs, which are included in:
        ``equipment_costs``.

        Parameters
        ----------
        frequency : str
            One of "project" or "annual", "monthly", ".

        Returns
        -------
        pd.DataFrame
            The broken out by time port fees with

        Raises
        ------
        ValueError
            If ``frequency`` not one of "project" or "annual".
        """
        frequency = _check_frequency(frequency, which="all")

        column = "port_fees"
        port_fee = self.events.loc[
            self.events.action == "monthly lease fee",
            ["year", "month", "equipment_cost"],
        ].rename(columns={"equipment_cost": column})

        if port_fee.shape[0] == 0:
            return pd.DataFrame([[0]], columns=[column])

        if frequency == "project":
            return pd.DataFrame([port_fee.sum(axis=0).loc[column]], columns=[column])
        elif frequency == "annual":
            return port_fee[["year"] + [column]].groupby(["year"]).sum()
        elif frequency == "monthly":
            return port_fee[["month"] + [column]].groupby(["month"]).sum()
        elif frequency == "month-year":
            return (
                port_fee[["year", "month"] + [column]].groupby(["year", "month"]).sum()
            )

    def project_fixed_costs(self, frequency: str, resolution: str) -> pd.DataFrame:
        """Calculates the fixed costs of a project at the project and annual frequencies
        at a given cost breakdown resolution.

        Parameters
        ----------
        frequency : str
            One of "project" or "annual", "monthly", ".
        resolution : st
            One of "low", "medium", or "high", where the values correspond to:

            - low: ``FixedCosts.resolution["low"]``, corresponding to itemized costs.
            - medium: ``FixedCosts.resolution["medium"]``, corresponding to the
              overarching cost categories.
            - high: ``FixedCosts.resolution["high"]``, corresponding to a lump sum.

            These values can also be seen through the ``FixedCosts.hierarchy``

        Returns
        -------
        pd.DataFrame
            The project's fixed costs as a sum or annualized with high, medium, and low
            resolution as desired.

        Raises
        ------
        ValueError
            If ``frequency`` not one of "project" or "annual".
        ValueError
            If ``resolution`` must be one of "low", "medium", or "high".
        """
        frequency = _check_frequency(frequency, which="all")

        resolution = resolution.lower().strip()
        if resolution not in ("low", "medium", "high"):
            raise ValueError(
                '``resolution`` must be one of "low", "medium", or "high".'
            )

        # Get the appropriate values and convert to the currency base
        keys = self.fixed_costs.resolution[resolution]
        vals = (
            np.array([[getattr(self.fixed_costs, key) for key in keys]])
            * self.project_capacity
            * 1000
        )

        total = (
            self.operations[["year", "month", "env_time"]]
            .groupby(["year", "month"])
            .count()
        )
        total = total.rename(columns={"env_time": "N"})
        total.N = 1.0

        operation_hours = (
            self.operations[["year", "month", "env_time"]]
            .groupby(["year", "month"])
            .count()
        )
        operation_hours = operation_hours.rename(columns={"env_time": "N"})

        costs = pd.DataFrame(total.values * vals, index=total.index, columns=keys)
        costs *= operation_hours.values.reshape(-1, 1) / 8760.0

        adjusted_inflation = np.array(
            [self.inflation_rate ** (i // 12) for i in range(costs.shape[0])]
        )
        costs *= adjusted_inflation.reshape(-1, 1)

        if frequency == "project":
            costs = pd.DataFrame(costs.reset_index(drop=True).sum()).T
        elif frequency == "annual":
            costs = costs.reset_index().groupby("year").sum().drop(columns=["month"])
        elif frequency == "monthly":
            costs = costs.reset_index().groupby("month").sum().drop(columns=["year"])

        return costs

    def opex(self, frequency: str) -> pd.DataFrame:
        """Calculates the project's OpEx for the simulation at a project, annual, or
        monthly level.

        Parameters
        ----------
        frequency : str
            One of project, annual, monthly, or month-year.

        Returns
        -------
        pd.DataFrame
            The project's OpEx broken out at the desired time resolution.
        """
        frequency = _check_frequency(frequency, which="all")

        # Get the materials costs and remove the component-level breakdown
        materials = self.component_costs(frequency=frequency, by_category=True)
        materials = materials.loc[:, ["materials_cost"]].reset_index()
        if frequency == "project":
            materials = pd.DataFrame(materials.loc[:, ["materials_cost"]].sum()).T
        else:
            if frequency == "annual":
                group_col = ["year"]
            elif frequency == "monthly":
                group_col = ["month"]
            elif frequency == "month-year":
                group_col = ["year", "month"]
            materials = (
                materials[group_col + ["materials_cost"]].groupby(group_col).sum()
            )

        # Port fees will produce an 1x1 dataframe if values aren't present, so recreate
        # it with the appropriate dimension
        port_fees = self.port_fees(frequency=frequency)
        if frequency != "project" and port_fees.shape == (1, 1):
            port_fees = pd.DataFrame([], columns=["port_fees"], index=materials.index)
            port_fees = port_fees.fillna(0)

        # Create a list of data frames for the OpEx components
        opex_items = [
            self.project_fixed_costs(frequency=frequency, resolution="low"),
            port_fees,
            self.equipment_costs(frequency=frequency),
            self.labor_costs(frequency=frequency),
            materials,
        ]

        # Join the data frames and sum along the time axis and return
        column = "OpEx"
        opex = pd.concat(opex_items, axis=1)
        opex.loc[:, column] = opex.sum(axis=1)
        return opex[[column]]

    def process_times(self) -> pd.DataFrame:
        """Calculates the time, in hours, to complete a repair/maintenance request, on
        both a request to completion basis, and the actual time to complete the repair.

        Returns
        -------
        pd.DataFrame
         - category (index): repair/maintenance category
         - time_to_completion: total number of hours from the time of request to the
           time of completion
         - process_time: total number of hours it took for the equipment to complete
         - the request.
         - downtime: total number of hours where the operations were below 100%.
         - N: total number of processes in the category.
        """
        events_valid = self.events.loc[self.events.request_id != "na"]

        # Summarize all the requests data
        request_df = (
            events_valid[["request_id", "env_time", "duration"]]
            .groupby("request_id")
            .sum()
            .sort_index()
        )
        request_df_min = (
            events_valid[["request_id", "env_time", "duration"]]
            .groupby("request_id")
            .min()
            .sort_index()
        )
        request_df_max = (
            events_valid[["request_id", "env_time", "duration"]]
            .groupby("request_id")
            .max()
            .sort_index()
        )

        # Summarize all the downtime-specific data for all requests
        downtime_df = events_valid.loc[events_valid.system_operating_level < 1][
            ["request_id", "env_time", "duration"]
        ]
        downtime_df_min = (
            downtime_df[["request_id", "env_time", "duration"]]
            .groupby("request_id")
            .min()
            .sort_index()
        )
        downtime_df_max = (
            downtime_df[["request_id", "env_time", "duration"]]
            .groupby("request_id")
            .max()
            .sort_index()
        )

        reason_df = (
            events_valid.drop_duplicates(subset=["request_id"])[
                ["request_id", "reason"]
            ]
            .set_index("request_id")
            .sort_index()
        )

        # Create the timing dataframe
        timing = pd.DataFrame([], index=request_df_min.index)
        timing = timing.join(reason_df[["reason"]]).rename(
            columns={"reason": "category"}
        )
        timing = timing.join(
            request_df_min[["env_time"]]
            .join(request_df_max[["env_time"]], lsuffix="_min", rsuffix="_max")
            .diff(axis=1)[["env_time_max"]]
            .rename(columns={"env_time_max": "time_to_completion"})
        )
        timing = timing.join(request_df[["duration"]]).rename(
            columns={"duration": "process_time"}
        )
        timing = timing.join(
            downtime_df_min[["env_time"]]
            .join(downtime_df_max[["env_time"]], lsuffix="_min", rsuffix="_max")
            .diff(axis=1)[["env_time_max"]]
            .rename(columns={"env_time_max": "downtime"})
        )
        timing["N"] = 1

        # Return only the categorically summed data
        return timing.groupby("category").sum().sort_index()

    def power_production(
        self, frequency: str, by: str = "windfarm", units: str = "gwh"
    ) -> float | pd.DataFrame:
        """Calculates the power production for the simulation at a project, annual, or
        monthly level that can be broken out by turbine.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        by : str
            One of "windfarm" or "turbine".
        units : str
            One of "gwh", "mwh", or "kwh".

        Returns
        -------
        float | pd.DataFrame
            Returns either a float for whole project-level costs or a pandas
            ``DataFrame`` with columns:

            - year (if appropriate for frequency)
            - month (if appropriate for frequency)
            - total_power_production
            - <turbine_id>_power_production (if broken out)

        Raises
        ------
        ValueError
            If ``frequency`` is not one of "project", "annual", "monthly", or
            "month-year".
        ValueError
            If ``by_turbine`` is not one of ``True`` or ``False``.
        """
        frequency = _check_frequency(frequency, which="all")

        by = by.lower().strip()
        if by not in ("windfarm", "turbine"):
            raise ValueError('``by`` must be one of "windfarm" or "turbine".')
        by_turbine = by == "turbine"

        if units not in ("gwh", "mwh", "kwh"):
            raise ValueError('``units`` must be one of "gwh", "mwh", or "kwh".')
        if units == "gwh":
            divisor = 1e6
            label = "Project Energy Production (GWh)"
        elif units == "mwh":
            divisor = 1e3
            label = "Project Energy Production (MWh)"
        else:
            divisor = 1
            label = "Project Energy Production (kWh)"

        if frequency == "annual":
            group_cols = ["year"]
        elif frequency == "monthly":
            group_cols = ["month"]
        elif frequency == "month-year":
            group_cols = ["year", "month"]

        col_filter = ["windfarm"]
        if by_turbine:
            col_filter.extend(self.turbine_id)

        if frequency == "project":
            production = self.production[col_filter].sum(axis=0)
            production = (
                pd.DataFrame(
                    production.values.reshape(1, -1),
                    columns=col_filter,
                    index=[label],
                )
                / divisor
            )
            return production
        return (
            self.production[group_cols + col_filter].groupby(by=group_cols).sum()
            / divisor
        )

    # Windfarm Financials

    def npv(
        self, frequency: str, discount_rate: float = 0.025, offtake_price: float = 80
    ) -> pd.DataFrame:
        """Calculates the net present value of the windfarm at a project, annual, or
        monthly resolution given a base discount rate and offtake price.

        .. note:: This function will be improved over time to incorporate more of the
            financial parameter at play, such as PPAs.

        Parameters
        ----------
        frequency : str
            One of "project", "annual", "monthly", or "month-year".
        discount_rate : float, optional
            The rate of return that could be earned on alternative investments, by
            default 0.025.
        offtake_price : float, optional
            Price of energy, per MWh, by default 80.

        Returns
        -------
        pd.DataFrame
            The project net prsent value at the desired time resolution.
        """
        frequency = _check_frequency(frequency, which="all")

        # Gather the OpEx, and revenues
        expenditures = self.opex("month-year")
        production = self.power_production("month-year")
        revenue: pd.DataFrame = production / 1000 * offtake_price  # MWh

        # Instantiate the NPV with the required calculated data and compute the result
        npv = revenue.join(expenditures).rename(columns={"windfarm": "revenue"})
        N = npv.shape[0]
        npv.loc[:, "discount"] = np.full(N, 1 + discount_rate) ** np.arange(N)
        npv.loc[:, "NPV"] = (npv.revenue.values - npv.OpEx.values) / npv.discount.values

        # Aggregate the results to the required resolution
        if frequency == "project":
            return pd.DataFrame(npv.reset_index().sum()).T[["NPV"]]
        elif frequency == "annual":
            return npv.reset_index().groupby("year").sum()[["NPV"]]
        elif frequency == "monthly":
            return npv.reset_index().groupby("month").sum()[["NPV"]]
        return npv[["NPV"]]

    def pysam_npv(self) -> float | pd.DataFrame:
        """Returns the project-level after-tax net present values (NPV).

        See here for more:
        https://nrel-pysam.readthedocs.io/en/master/modules/Singleowner.html#PySAM.Singleowner.Singleowner.Outputs.cf_project_return_aftertax_npv

        .. warning::
            PySAM functionality is currently disabled due to changes made in the new API
            that need to be remapped.

        Raises
        ------
        NotImplementedError: Raised if a PySAM input file is not provided to run the
        model.

        Returns
        -------
        float
            Final, project-level NPV, in $.
        """
        if self.financial_model is None:
            raise NotImplementedError(
                "No SAM inputs were provided, and 'pysam_npv()' cannot be calculated!"
            )
        npv = self.financial_model.Outputs.cf_project_return_aftertax_npv
        npv = npv[len(self.years)]
        return npv

    def pysam_lcoe_real(self) -> float:
        """Returns the real levelized cost of energy (LCOE) from PySAM.

        See here for more:
        https://nrel-pysam.readthedocs.io/en/master/modules/Singleowner.html#PySAM.Singleowner.Singleowner.Outputs.lcoe_real

        .. warning::
            PySAM functionality is currently disabled due to changes made in the new API
            that need to be remapped.

        Raises
        ------
        NotImplementedError: Raised if a PySAM input file is not provided to run the
            model.

        Returns
        -------
        float
            Real LCOE, in $/kW.
        """
        if self.financial_model is None:
            raise NotImplementedError(
                "No SAM inputs were provided, and 'pysam_lcoe_real()' cannot be"
                " calculated!"
            )
        return self.financial_model.Outputs.lcoe_real / 100.0

    def pysam_lcoe_nominal(self) -> float:
        """Returns the nominal levelized cost of energy (LCOE) from PySAM.

        See here for more:
        https://nrel-pysam.readthedocs.io/en/master/modules/Singleowner.html#PySAM.Singleowner.Singleowner.Outputs.lcoe_nom

        .. warning::
            PySAM functionality is currently disabled due to changes made in the new API
            that need to be remapped.

        Raises
        ------
        NotImplementedError: Raised if a PySAM input file is not provided to run the
        model.

        Returns
        -------
        float
            Nominal LCOE, in $/kW.
        """
        if self.financial_model is None:
            raise NotImplementedError(
                "No SAM inputs were provided, and 'pysam_lcoe_nominal()' cannot"
                " be calculated!"
            )
        return self.financial_model.Outputs.lcoe_nom / 100.0

    def pysam_irr(self) -> float:
        """Returns the project-level after-tax internal return rate (IRR).

        See here for more:
        https://nrel-pysam.readthedocs.io/en/master/modules/Singleowner.html#PySAM.Singleowner.Singleowner.Outputs.cf_project_return_aftertax_irr

        .. warning::
            PySAM functionality is currently disabled due to changes made in the new API
            that need to be remapped.

        Raises
        ------
        NotImplementedError: Raised if a PySAM input file is not provided to run the
            model.

        Returns
        -------
        pd.DataFrame
            Annual after-tax IRR value, in %.
        """
        if self.financial_model is None:
            raise NotImplementedError(
                "No SAM inputs were provided, and 'pysam_irr()' cannot be calculated!"
            )
        irr = self.financial_model.Outputs.cf_project_return_aftertax_irr
        irr = irr[len(self.years)]
        return irr

    def pysam_all_outputs(self) -> pd.DataFrame:
        """Returns all the possible PySAM outputs that are included in this module as
        columns in the following order.

        - NPV
        - Nominal LCOE
        - Real LOCE
        - IRR

        .. warning::
           PySAM functionality is currently disabled due to changes made in the new API
           that need to be remapped.

        Raises
        ------
        NotImplementedError: Raised if a PySAM input file is not provided to run the
            model.

        Returns
        -------
        pd.DataFrame
            Project financial values values.
        """
        if self.financial_model is None:
            raise NotImplementedError(
                "No SAM inputs were provided, and 'pysam_all_outputs()' cannot"
                " be calculated!"
            )
        financials = [
            self.pysam_npv(),
            self.pysam_lcoe_nominal(),
            self.pysam_lcoe_real(),
            self.pysam_irr(),
        ]
        descriptions = [
            "After Tax NPV ($)",
            "Nominal LCOE ($/kW)",
            "Real LCOE ($/kW)",
            "After Tax IRR (%)",
        ]
        financials = pd.DataFrame(
            financials, index=descriptions, dtype=float, columns=["Value"]
        )
        financials.index.name = "Metric"  # type: ignore
        return financials



"""Provides various utility functions that don't fit within a common theme."""


from __future__ import annotations

import re
from typing import Callable

import numpy as np
import pandas as pd


try:
    from functools import cache  # type: ignore
except ImportError:
    from functools import lru_cache

    cache = lru_cache(None)


# Don't repeat the most common inputs that occur when there are no state changes, but
# the result is needed again for logging
@cache
def _mean(*args) -> float:
    """Multiplies two numbers. Used for a reduce operation.

    Parameters
    ----------
    args : int | float
        The values to compute the mean over

    Returns
    -------
    float
        The average of the values provided
    """
    return np.mean(args)


def create_variable_from_string(string: str) -> str:
    """Creates a valid Python variable style string from a passed string.

    Parameters
    ----------
    string : str
        The string to convert into a Python-friendly variable name.

    Returns
    -------
    str
        A Python-valid variable name.

    Examples
    --------
    >>> example_string = "*Electrical!*_ _System$*_"
    >>> print(create_variable_from_string(example_string))
    'electrical_system'

    """
    new_string = re.sub(
        "[^0-9a-zA-Z]+",
        "_",
        re.sub("^\W+", "", re.sub("[^a-zA-Z]+$", "", string)),  # noqa: disable=W605
    ).lower()
    return new_string


def IEC_power_curve(
    windspeed_column: np.ndarray | pd.Series,
    power_column: np.ndarray | pd.Series,
    bin_width: float = 0.5,
    windspeed_start: float = 0.0,
    windspeed_end: float = 30.0,
) -> Callable:
    """Direct copyfrom OpenOA:
    https://github.com/NREL/OpenOA/blob/main/operational_analysis/toolkits/power_curve/functions.py#L16-L57
    Use IEC 61400-12-1-2 method for creating wind-speed binned power curve.

    Parameters
    ----------
        windspeed_column : np.ndarray | pandas.Series
            The power curve's windspeed values, in m/s.
        power_column : np.ndarray | pandas.Series
            The power curve's output power values, in kW.
        bin_width : float
            Width of windspeed bin, default is 0.5 m/s according to standard, by default
            0.5.
        windspeed_start : float
            Left edge of first windspeed bin, where all proceeding values will be 0.0,
            by default 0.0.
        windspeed_end : float
            Right edge of last windspeed bin, where all following values will be 0.0, by
            default 30.0.

    Returns
    -------
        Callable
            Python function of the power curve, of type (Array[float] -> Array[float]),
            that maps input windspeed value(s) to ouptut power value(s).
    """
    if not isinstance(windspeed_column, pd.Series):
        windspeed_column = pd.Series(windspeed_column)
    if not isinstance(power_column, pd.Series):
        power_column = pd.Series(power_column)

    # Set up evenly spaced bins of fixed width, with np.inf for any value over the max
    n_bins = int(np.ceil((windspeed_end - windspeed_start) / bin_width)) + 1
    bins = np.append(np.linspace(windspeed_start, windspeed_end, n_bins), [np.inf])

    # Initialize an array which will hold the mean values of each bin
    P_bin = np.ones(len(bins) - 1) * np.nan

    # Compute the mean of each bin and set corresponding P_bin
    for ibin in range(0, len(bins) - 1):
        indices = (windspeed_column >= bins[ibin]) & (windspeed_column < bins[ibin + 1])
        P_bin[ibin] = power_column.loc[indices].mean()

    # Linearly interpolate any missing bins
    P_bin = pd.Series(data=P_bin).interpolate(method="linear").bfill().values

    # Create a closure over the computed bins which computes the power curve value for
    # arbitrary array-like input
    def pc_iec(x):
        P = np.zeros(np.shape(x))
        for i in range(0, len(bins) - 1):
            idx = np.where((x >= bins[i]) & (x < bins[i + 1]))
            P[idx] = P_bin[i]
        cutoff_idx = (x < windspeed_start) | (x > windspeed_end)
        P[cutoff_idx] = 0.0
        return P

    return pc_iec



"""Test the Cable system and subassembly classes."""


from wombat.core import RepairManager
from wombat.windfarm import Windfarm
from wombat.windfarm.system import Cable

from tests.conftest import ARRAY_33KV_240MM, ARRAY_33KV_630MM


def test_cable_init(env_setup):
    """Tests the initialization of a `Cable` object. Much of this exists in the windfarm
    tests, but is repeated here for the sake of redundnacy and clarity that tests exist.
    """
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout.csv", manager)

    correct_edge_keys = [
        ("OSS1", "S00T1"),
        ("S00T1", "S00T2"),
        ("S00T2", "S00T3"),  # string 1
        ("OSS1", "S01T4"),
        ("S01T4", "S01T5"),
        ("S01T5", "S01T6"),  # string 2
    ]
    correct_cable_ids = ["::".join(("cable", *el)) for el in correct_edge_keys]
    correct_cable_list = [
        Cable(windfarm, env, "array", "OSS1", "S00T1", ARRAY_33KV_630MM),
        Cable(windfarm, env, "array", "S00T1", "S00T2", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "S00T2", "S00T3", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "OSS1", "S01T4", ARRAY_33KV_630MM),
        Cable(windfarm, env, "array", "S01T4", "S01T5", ARRAY_33KV_240MM),
        Cable(windfarm, env, "array", "S01T5", "S01T6", ARRAY_33KV_240MM),
    ]

    for key, _id, correct_cable in zip(
        correct_edge_keys, correct_cable_ids, correct_cable_list
    ):
        cable = windfarm.graph.edges[key]["cable"]
        assert cable.env == correct_cable.env
        assert cable.id == correct_cable.id
        assert cable.name == correct_cable.name
        assert cable.system == correct_cable.system
        assert cable.data.name == cable.data.name
        assert cable.operating_level == correct_cable.operating_level == 1
        assert cable.broken.triggered
        assert cable.servicing.triggered
        assert cable.downstream_failure.triggered

        # Rough check of the subassemblies and maintenance/failure creation
        assert cable.processes.keys() == correct_cable.processes.keys()


def test_cable_failures(env_setup):
    """Test that failing cable disable upstream turbines."""
    env = env_setup
    manager = RepairManager(env)
    windfarm = Windfarm(env, "layout_cable_test.csv", manager)

    # env.run(1)
    # from pprint import pprint
    # for name, cable in windfarm.graph.edges.items():
    #     if name != ("OSS1", "S00T1"):
    #         continue
    #     cable = cable["cable"]
    #     print(name)
    #     for pname, p in cable.processes.items():
    #         print(pname)
    #         pprint(p.__dict__)

    # A failure will occur at 13.85 for the cable connecting the OSS to the first
    # turbine, so check that all turbines are down, and this should effectively shut
    # down the farm.
    catastrophic_timeout = 13.85
    env.run(catastrophic_timeout + 1)
    assert windfarm.current_availability == 0.5  # 50% of the farm is off

    # Check the cable is broken
    cable = windfarm.cable(("OSS1", "S00T1"))
    assert cable.broken
    assert cable.operating_level == 0

    # Assert that all items that are timed are non existent or set to the very end
    # for p in cable.processes.values():
    #     pprint(p._target.__dict__)
    assert getattr(list(cable.processes.values())[0]._target, "_delay", None) is None
    assert (
        getattr(list(cable.processes.values())[1]._target, "_delay", None)
        == 1402.148783264902
    )

    # Check the failure was submitted and no other items exist for this cable
    assert sum(req.system_id == "cable::OSS1::S00T1" for req in manager.items) == 1

    # Check the substation is operating (no downstream nodes are shut off)
    oss = windfarm.system("OSS1")
    assert oss.cable_failure.triggered
    assert oss.operating_level == 1

    # Check the appropriate turbines are broken
    for turbine in ("S00T1", "S00T2", "S00T3"):
        turbine = windfarm.system(turbine)
        assert not turbine.cable_failure.triggered
        assert turbine.operating_level == 0

        # Check that operating level for subassemblies is unaffected, and 0 turbine
        # operations is controlled by cable failure flag
        assert all(s.operating_level == 1 for s in turbine.subassemblies)

    # Check the appropriate turbines are operating
    for turbine in ("S01T4", "S01T5", "S01T6"):
        turbine = windfarm.system(turbine)
        assert turbine.cable_failure.triggered
        assert turbine.operating_level == 1
        assert all(s.operating_level == 1 for s in turbine.subassemblies)
